/*asm.c 80386 assembler output of same form as C compiler*/
#include <stdio.h>
#include <ctype.h>
#define NOFUNCTIONS
#include "pseudo.h" /* for OFUNC,OSTATFUNC,OPTEXTERN,& Stypes */

#define MAXINSZ  16 /*instruction name size*/
#define MAXMINDX 16 /*if max. instruction length is 15... */
#define MAXARGS  8
#define STRINGSZ 256
#define BUFSZ    32 /* mem. allocated in chunks this big */

/* following are used with memory type instructions */
#define O8   0x1 /* 8-bit operands OK */
#define O16  0x2 /* 16-bit operands OK */
#define O32  0x4 /* 32-bit operands OK */
#define RG   0x8 /* reg for mem OK */
#define ONONE 0x10 /* not 16/32, either should not be interpreted */
#define OBT  (O16|O32) /* both types OK */
#define OAL  (RG|OBT) /* 16/32 bit and regs OK */

#define MFLT 0 /* math float operands */
#define MINT 1 /* math integer operands */

#define E_MEM -1

#define AL(place,size,type) { (place)=(type *)malloc((unsigned)(sizeof(type)\
  *(size))+1);\
 if ((char *)(place)==NULL)\
  fatalerror("AL() can't get enough memory"),exit(E_MEM); }
#define RAL(place,size,type) { (place)=(type *)realloc((char *)(place)\
  ,(unsigned)(sizeof(type)*(size))+1);\
 if ((char *)(place)==NULL)\
  fatalerror("RAL() can't get enough memory"),exit(E_MEM); }
char *malloc(),*realloc();

long getnum();
int find_non_number();
int byte(),dblbyte(),general(),memreg(),regmem2(),bittest();
int transfer(),copbyte(),incdec(),muldiv(),enter(),imul(),inout(),interrupt();
int jcc(),memory2(),loop(),mov(),movxt(),negnot(),pushpop(),shftrot();
int dblshift(),test(),xchg(),modes(),returnproc();
int datab(),dataw(),absvardef(),funcinit(),endfunc(),stackvars(),stackptr();
int extvardef(),datadw(),escape(),fgenops(),fintern(),fst_ld();
int memory();

typedef int varcnttp,outoff;
typedef unsigned short *wordptr;
typedef unsigned long *dwordptr;
/* pseudo instructions:
 *modes [16bit/32bit] [expsz/noexpsz]
 *modes [expseg/noexpseg] [86/286/386] [jccnear/nojccnear]
 *absolute ID,ID,ID
 *func name tp sz(only if struct or ptr)
 *stackvars name sz name sz...
 *stackptr
 */
#define TRCLL  1  /* tell transfer() what type of instruction*/
#define TRJMP  2
struct instructions
{ char nm[MAXINSZ];
  int (*func)(),arg[MAXARGS];
} iset[]=
 {
  {"aaa"   ,byte    ,0x37,0   },
  {"aad"   ,dblbyte ,0xd5,0x0a},
  {"aam"   ,dblbyte ,0xd4,0x0a},
  {"aas"   ,byte    ,0x3f,0   },
  {"absolute",absvardef       },
  {"adc"   ,general ,0x10     },
  {"add"   ,general ,0x0      },
  {"and"   ,general ,0x20     },
  {"arpl"  ,memreg  ,0x63,O16|RG,O16},
  {"bound" ,regmem2 ,0,0,0x62,OBT   ,OBT },
  {"bsf"   ,regmem2 ,1,0x0f,0xbc  ,OBT,OAL},
  {"bsr"   ,regmem2 ,1,0x0f,0xbd  ,OBT,OAL},
  {"bt"    ,bittest ,0xa3     },
  {"btc"   ,bittest ,0xbb     },
  {"btr"   ,bittest ,0xb3     },
  {"bts"   ,bittest ,0xab     },
  {"call"  ,transfer,TRCLL,0xe8},
  {"cbw"   ,copbyte ,0x98,O16,0},
  {"cdq"   ,copbyte ,0x99,O32,0},
  {"clc"   ,byte    ,0xf8,0   },
  {"cld"   ,byte    ,0xfc,0   },
  {"cli"   ,byte    ,0xfa,0   },
  {"clts"  ,dblbyte ,0x0f,0x06},
  {"cmc"   ,byte    ,0xf5,0   },
  {"cmp"   ,general ,0x38     },
  {"cmpsb" ,byte    ,0xa6,1   },
  {"cmpsd" ,copbyte ,0xa7,O32,1},
  {"cmpsw" ,copbyte ,0xa7,O16,1},
  {"cwd"   ,copbyte ,0x99,O16,0},
  {"cwde"  ,copbyte ,0x98,O32,0},
  {"daa"   ,byte    ,0x27,0   },
  {"das"   ,byte    ,0x2f,0   },
  {"db"    ,datab             },
  {"ddw"   ,datadw            },
  {"dec"   ,incdec  ,0x8      },
  {"div"   ,muldiv  ,0x30     },
  {"dw"    ,dataw             },
  {"endfunc",endfunc          },
  {"enter" ,enter             },
  {"esc"   ,escape            },
  {"extern",extvardef         },
  {"f2xm1" ,dblbyte ,0xd9,0xf0},
  {"fabs"  ,dblbyte ,0xd9,0xe1},
  {"fadd"  ,fgenops ,0x00,MFLT},
  {"faddp" ,fintern ,0xde,0xc0,1},
  {"fbld"  ,memory  ,0xdf,0x20,0,0},
  {"fbstp" ,memory  ,0xdf,0x30,0,0},
  {"fchs"  ,dblbyte ,0xd9,0xe0},
  {"fclex" ,dblbyte ,0xdb,0xe2},
  {"fcom"  ,fgenops ,0x10,MFLT},
  {"fcomp" ,fgenops ,0x18,MFLT},
  {"fcompp",dblbyte ,0xde,0xd9},
  {"fcos"  ,dblbyte ,0xd9,0xff},
  {"fdecstp",dblbyte,0xd9,0xf6},
  {"fdisi" ,dblbyte ,0xdb,0xe1}, /*286 specific*/
  {"fdiv"  ,fgenops ,0x30,MFLT},
  {"fdivp" ,fintern ,0xde,0xf8,1},
  {"fdivr" ,fgenops ,0x38,MFLT},
  {"fdivrp",fintern ,0xde,0xf0,1},
  {"feni"  ,dblbyte ,0xdb,0xe0},  /*286 specific*/
  {"ffree" ,fintern ,0xdd,0xf0,0},
  {"fiadd" ,fgenops ,0x00,MINT},
  {"ficom" ,fgenops ,0x10,MINT},
  {"ficomp",fgenops ,0x18,MINT},
  {"fidiv" ,fgenops ,0x30,MINT},
  {"fidivr",fgenops ,0x38,MINT},
  {"fild"  ,fst_ld  ,0x0 ,MINT},
  {"fimul" ,fgenops ,0x08,MINT},
  {"fincstp",dblbyte,0xd9,0xf7},
  {"finit" ,dblbyte ,0xdb,0xe3},
  {"fist"  ,fst_ld  ,0x10,MINT},
  {"fistp" ,fst_ld  ,0x18,MINT},
  {"fisub" ,fgenops ,0x20,MINT},
  {"fisubr",fgenops ,0x28,MINT},
  {"fld"   ,fst_ld  ,0x0 ,MFLT},
  {"fld1"  ,dblbyte ,0xd9,0xe8},
  {"fldcw" ,memory  ,0xd9,0x28,0,0},
  {"fldenv",memory  ,0xd9,0x20,1,0},
  {"fldl2e",dblbyte ,0xd9,0xea},
  {"fldl2t",dblbyte ,0xd9,0xe9},
  {"fldlg2",dblbyte ,0xd9,0xec},
  {"fldln2",dblbyte ,0xd9,0xed},
  {"fldpi" ,dblbyte ,0xd9,0xeb},
  {"fldz"  ,dblbyte ,0xd9,0xee},
  {"fmul"  ,fgenops ,0x08,MFLT},
  {"fmulp" ,fintern ,0xde,0xc8,1},
  {"fnop"  ,dblbyte ,0xd9,0x00},
  {"fpatan",dblbyte ,0xd9,0xf3},
  {"fprem" ,dblbyte ,0xd9,0xf8},
  {"fprem1",dblbyte ,0xd9,0xf5},
  {"fptan" ,dblbyte ,0xd9,0xf2},
  {"frndint",dblbyte,0xd9,0xfc},
  {"frstor",memory  ,0xdd,0x20,1,0},
  {"fsave" ,memory  ,0xdd,0x30,1,0},
  {"fscale",dblbyte ,0xd9,0xfd},
  {"fsetpm",dblbyte ,0xdb,0xe4},    /*286 specific*/
  {"fsin"  ,dblbyte ,0xd9,0xfe},
  {"fsincos",dblbyte,0xd9,0xfb},
  {"fsqrt" ,dblbyte ,0xd9,0xfa},
  {"fst"   ,fst_ld  ,0x10,MFLT},
  {"fstcw" ,memory  ,0xd9,0x38,0,0},
  {"fstenv",memory  ,0xd9,0x30,1,0},
  {"fstp"  ,fst_ld  ,0x18,MFLT},
  {"fstsw" ,memory  ,0xdd,0x38,0,1},
  {"fsub"  ,fgenops ,0x20,MFLT},
  {"fsubp" ,fintern ,0xde,0xe8,1},
  {"fsubr" ,fgenops ,0x28,MFLT},
  {"fsubrp",fintern ,0xde,0xe0,1},
  {"ftst"  ,dblbyte ,0xd9,0xe4},
  {"fucom" ,fintern ,0xdd,0xe0,0},
  {"fucomp",fintern ,0xdd,0xe8,0},
  {"fucompp",dblbyte,0xda,0xe9},
  {"func"  ,funcinit,OFUNC    },
  {"fxam"  ,dblbyte ,0xd9,0xe5},
  {"fxch"  ,fintern ,0xd9,0xc8,1},
  {"fxtract",dblbyte,0xd9,0xf4},
  {"fyl2x" ,dblbyte ,0xd9,0xf1},
  {"fyl2xp1",dblbyte,0xd9,0xf9},
  {"hlt"   ,byte    ,0xf4,0   },
  {"idiv"  ,muldiv  ,0x38     },
  {"imul"  ,imul              },
  {"in"    ,inout   ,0xe4     },
  {"inc"   ,incdec  ,0x0      },
  {"insb"  ,byte    ,0x6c,1   },
  {"insd"  ,copbyte ,0x6d,O32,1},
  {"insw"  ,copbyte ,0x6d,O16,1},
  {"int"   ,interrupt         },
  {"into"  ,byte    ,0xce,0   },
  {"iret"  ,byte    ,0xcf,0   },
  {"iretd" ,copbyte ,0xcf,O32,0},
  {"iretw" ,copbyte ,0xcf,O16,0},
  {"ja"    ,jcc     ,0x77,0   },
  {"jae"   ,jcc     ,0x73,0   },
  {"jb"    ,jcc     ,0x72,0   },
  {"jbe"   ,jcc     ,0x76,0   },
  {"jc"    ,jcc     ,0x72,0   },
  {"jcxz"  ,jcc     ,0xe3,1   },
  {"je"    ,jcc     ,0x74,0   },
  {"jecxz" ,jcc     ,0xe3,1   },
  {"jg"    ,jcc     ,0x7f,0   },
  {"jge"   ,jcc     ,0x7d,0   },
  {"jl"    ,jcc     ,0x7c,0   },
  {"jle"   ,jcc     ,0x7e,0   },
  {"jmp"   ,transfer,TRJMP,0xeb},
  {"jna"   ,jcc     ,0x76,0   },
  {"jnae"  ,jcc     ,0x72,0   },
  {"jnb"   ,jcc     ,0x73,0   },
  {"jnbe"  ,jcc     ,0x77,0   },
  {"jnc"   ,jcc     ,0x73,0   },
  {"jne"   ,jcc     ,0x75,0   },
  {"jng"   ,jcc     ,0x7e,0   },
  {"jnge"  ,jcc     ,0x7c,0   },
  {"jnl"   ,jcc     ,0x7d,0   },
  {"jnle"  ,jcc     ,0x7f,0   },
  {"jno"   ,jcc     ,0x71,0   },
  {"jnp"   ,jcc     ,0x7b,0   },
  {"jns"   ,jcc     ,0x79,0   },
  {"jnz"   ,jcc     ,0x75,0   },
  {"jo"    ,jcc     ,0x70,0   },
  {"jp"    ,jcc     ,0x7a,0   },
  {"jpe"   ,jcc     ,0x7a,0   },
  {"jpo"   ,jcc     ,0x7b,0   },
  {"js"    ,jcc     ,0x78,0   },
  {"jz"    ,jcc     ,0x74,0   },
  {"lahf"  ,byte    ,0x9f,0   },
  {"lar"   ,regmem2 ,1,0x0f,0x02,OBT,OAL},
  {"lds"   ,regmem2 ,0,0,0xc5,OBT ,OBT},
  {"lea"   ,regmem2 ,0,0,0x8d,OBT ,ONONE},
  {"leave" ,byte    ,0xc9,0   },
  {"les"   ,regmem2 ,0,0,0xc4,OBT ,OBT},
  {"lfs"   ,regmem2 ,1,0x0f,0xb4,OBT ,OBT},
  {"lgdt"  ,memory2 ,0x0f,0x01,0x10,OBT},
  {"lgs"   ,regmem2 ,1,0x0f,0xb5,OBT ,OBT},
  {"lidt"  ,memory2 ,0x0f,0x01,0x18,OBT},
  {"lldt"  ,memory2 ,0x0f,0x00,0x10,O16},
  {"lmsw"  ,memory2 ,0x0f,0x01,0x30,O16},
  {"lodsb" ,byte    ,0xac,1   },
  {"lodsd" ,copbyte ,0xad,O32,1},
  {"lodsw" ,copbyte ,0xad,O16,1},
  {"loop"  ,loop    ,0xe2     },
  {"loope" ,loop    ,0xe1     },
  {"loopne",loop    ,0xe0     },
  {"loopnz",loop    ,0xe0     },
  {"loopz" ,loop    ,0xe1     },
  {"lsl"   ,regmem2 ,1,0x0f,0x03,OBT ,OAL},
  {"lss"   ,regmem2 ,1,0x0f,0xb2,OBT ,OBT},
  {"ltr"   ,memory2 ,0x0f,0x00,0x18,O16},
  {"modes" ,modes             },
  {"mov"   ,mov               },
  {"movsb" ,byte    ,0xa4,1   },
  {"movsd" ,copbyte ,0xa5,O32,1},
  {"movsw" ,copbyte ,0xa5,O16,1},
  {"movsx" ,movxt   ,0xbe     },
  {"movzx" ,movxt   ,0xb6     },
  {"mul"   ,muldiv  ,0x20     },
  {"neg"   ,negnot  ,0x18     },
  {"nop"   ,byte    ,0x90,0   },
  {"not"   ,negnot  ,0x10     },
  {"or"    ,general ,0x08     },
  {"out"   ,inout   ,0xe6     },
  {"outsb" ,byte    ,0x6e,1   },
  {"outsd" ,copbyte ,0x6f,O32,1},
  {"outsw" ,copbyte ,0x6f,O16,1},
  {"pop"   ,pushpop ,0x08     },
  {"popa"  ,copbyte ,0x61,O16,0},
  {"popad" ,copbyte ,0x61,O32,0},
  {"popf"  ,copbyte ,0x9d,O16,0},
  {"popfd" ,copbyte ,0x9d,O32,0},
  {"push"  ,pushpop ,0x0      },
  {"pusha" ,copbyte ,0x60,O16,0},
  {"pushad",copbyte ,0x60,O32,0},
  {"pushf" ,copbyte ,0x9c,O16,0},
  {"pushfd",copbyte ,0x9c,O32,0},
  {"rcl"   ,shftrot ,0x10     },
  {"rcr"   ,shftrot ,0x18     },
  {"ret"   ,returnproc,0xc2   },
  {"retf"  ,returnproc,0xca   },
  {"rol"   ,shftrot ,0x0      },
  {"ror"   ,shftrot ,0x08     },
  {"sahf"  ,byte    ,0x9e,0   },
  {"sal"   ,shftrot ,0x20     },
  {"sar"   ,shftrot ,0x38     },
  {"sbb"   ,general ,0x18     },
  {"scasb" ,byte    ,0xae,1   },
  {"scasd" ,copbyte ,0xaf,O32,1},
  {"scasw" ,copbyte ,0xaf,O16,1},
  {"seta"  ,memory2 ,0x0f,0x97,0x0,O8|RG},
  {"setae" ,memory2 ,0x0f,0x93,0x0,O8|RG},
  {"setb"  ,memory2 ,0x0f,0x92,0x0,O8|RG},
  {"setbe" ,memory2 ,0x0f,0x96,0x0,O8|RG},
  {"setc"  ,memory2 ,0x0f,0x92,0x0,O8|RG},
  {"sete"  ,memory2 ,0x0f,0x94,0x0,O8|RG},
  {"setg"  ,memory2 ,0x0f,0x9f,0x0,O8|RG},
  {"setge" ,memory2 ,0x0f,0x9d,0x0,O8|RG},
  {"setl"  ,memory2 ,0x0f,0x9c,0x0,O8|RG},
  {"setle" ,memory2 ,0x0f,0x9e,0x0,O8|RG},
  {"setna" ,memory2 ,0x0f,0x96,0x0,O8|RG},
  {"setnae",memory2 ,0x0f,0x92,0x0,O8|RG},
  {"setnb" ,memory2 ,0x0f,0x93,0x0,O8|RG},
  {"setnbe",memory2 ,0x0f,0x97,0x0,O8|RG},
  {"setnc" ,memory2 ,0x0f,0x93,0x0,O8|RG},
  {"setne" ,memory2 ,0x0f,0x95,0x0,O8|RG},
  {"setng" ,memory2 ,0x0f,0x9e,0x0,O8|RG},
  {"setnge",memory2 ,0x0f,0x9c,0x0,O8|RG},
  {"setnl" ,memory2 ,0x0f,0x9d,0x0,O8|RG},
  {"setnle",memory2 ,0x0f,0x9f,0x0,O8|RG},
  {"setno" ,memory2 ,0x0f,0x91,0x0,O8|RG},
  {"setnp" ,memory2 ,0x0f,0x9b,0x0,O8|RG},
  {"setns" ,memory2 ,0x0f,0x99,0x0,O8|RG},
  {"setnz" ,memory2 ,0x0f,0x95,0x0,O8|RG},
  {"seto"  ,memory2 ,0x0f,0x90,0x0,O8|RG},
  {"setp"  ,memory2 ,0x0f,0x9a,0x0,O8|RG},
  {"setpe" ,memory2 ,0x0f,0x9a,0x0,O8|RG},
  {"setpo" ,memory2 ,0x0f,0x9b,0x0,O8|RG},
  {"sets"  ,memory2 ,0x0f,0x98,0x0,O8|RG},
  {"setz"  ,memory2 ,0x0f,0x94,0x0,O8|RG},
  {"sgdt"  ,memory2 ,0x0f,0x01,0x0,ONONE},
  {"shl"   ,shftrot ,0x20     },
  {"shld"  ,dblshift,0xa4     },
  {"shr"   ,shftrot ,0x28     },
  {"shrd"  ,dblshift,0xac     },
  {"sidt"  ,memory2 ,0x0f,0x01,0x08,ONONE},
  {"sldt"  ,memory2 ,0x0f,0x00,0x00,O16|RG},
  {"smsw"  ,memory2 ,0x0f,0x01,0x20,O16|RG},
  {"stackptr",stackptr        },
  {"stackvars",stackvars      },
  {"statfunc",funcinit,OSTATFUNC},
  {"stc"   ,byte    ,0xf9,0   },
  {"std"   ,byte    ,0xfd,0   },
  {"sti"   ,byte    ,0xfb,0   },
  {"stosb" ,byte    ,0xaa,1   },
  {"stosd" ,copbyte ,0xab,O32,1},
  {"stosw" ,copbyte ,0xab,O16,1},
  {"str"   ,memory2 ,0x0f,0x00,0x08,O16|RG},
  {"sub"   ,general ,0x28     },
  {"test"  ,test              },
  {"verr"  ,memory2 ,0x0f,0x00,0x20,O16|RG},
  {"verw"  ,memory2 ,0x0f,0x00,0x28,O16|RG},
  {"wait"  ,byte    ,0x9b,0   },
  {"xchg"  ,xchg              },
  {"xlat"  ,byte    ,0xd7,0   },
  {"xlatb" ,byte    ,0xd7,0   },
  {"xor"   ,general ,0x30     },
  {"",0}
 } ;
/*segments:auto&args in ss   HEREfuture
 *         public&static in ds
 *         segments of functions
 *         segs as part of pointers
 *         fixed-position segments  (no real counter for C)
 *         own defined segments of this program(no counter in C)
 */
struct externxref
{ long blck,off,scoff;
  char ptrfnc;
} *exref;
int exrcnt,lastextpres,lastextnum;

struct stackvars /* temp. store stack vars until stackptr encountered */
{ char *nm;
  outoff off; /* offset related to stkoff */
} *stkvars;
int stkvcnt,stkvAL,stkptrexist;
outoff stkoff,stkptr;

struct jloc /* jump location (also used with off, call, and jcc) */
{ char jloctp;
  varcnttp num;
  outoff off; /* offset from pos specified by u */
};
#define JLLABEL 1 /* jump location's type  is label*/
#define JLCONST 2
#define JLFUNC  3

struct varxref
{ int blk,line;
  char tp;
  varcnttp vnum; /* allocation num, not index num */
  outoff off;
  union /* type gotten from svars */
  { struct
    { struct jloc jloc;
      char opsz,jsz,isoff;
    } lab;
    struct
    { char addsz;
    } abs;
    struct
    { char ptrfnc,addsz;
    } ext;
  } u;
} *vxref;
int vxcnt;

struct exttbl
{ char *nm,tp,*prestf,prepres,scope;
  long presz,auxsz,sz;
} *exttbl;
int exttblsz,extxcnt;

#define SCPUBLIC 2    /* offset is offset of table being built */
struct prexref
{ char sect;
  long eoff,soff,poff;
} *extxref;
/* eoff-extern table index of variable being filled
 * sect,soff-sect/offset to ID the # placed in table(see #defines above)
 * poff-offset in variable where pointer is
 */

struct variables
/* stores necessary info on all variables, && labels of all types
 *  in a sorted list so that can be found with binary search */
{ char *nm,here,tp,local;
  varcnttp num; /* what number is used in xrefs; given in order encountered */
  union
  { struct
    { int blk;
      outoff off;
    } bp; /* blk position */
    outoff off;
  } u;
} *svars;
varcnttp *vars,varcnt;
#define VTNONE     0 /* used only by memindx */
#define VTLABEL    1 /*use bp; lab*/ /*use svars u; varxref u.*/
#define VTABSOLUTE 2 /*use bp; abs*/
#define VTSTACK    3 /*use off; none*/
#define VTEXTERN   5 /*use off; ext; off is offset in exttbl*/

struct outblock /* main output broken into blocks ending with jmps */
{ char *bops,jtp,jop,jsz,opsz; /*jop is short version of whatever jmp*/
  int line;
  struct jloc jloc;
  outoff blen,inoff;
} *blk;
int blkcnt;
#define JTJMP   1 /* jump types; jcxz is coded with xrefs */
#define JTJcc   2
#define JTLOOP  3
#define JTCALL  4
#define JTNONE  5 /* HEREfuture used with constant offsets and stuff */

struct prefix  /* used to store explicitly-specified prefixes */
{ int ipre,seg,opsz,addsz; } pfx;

struct memindx
{ char stf[MAXMINDX];
  int sz,xexist,xoff,addsz;
  int scoff;
  struct jloc jl;
}; /* xexist=VTsomething */

struct operand
{ int addsz,tp,reg,opsz,jsz,lastop,sign; /* sign only used with TLABEL. */
  int seg,tplab; /* seg=actual value of seg portion for call/jmp */
  long cval;
  struct memindx m;
  char string[STRINGSZ];
};
/* for jumps:
 * TCONST: just offset, no exact location
 * TMEM: tplab==1: lea xx,mem ; must generate from m.jl for jmp
 *            ==0: only place jmp [mem] OK
 * jsz will be SUNKNOWN/SFAR/SNEAR
 *  seg only defined if SFAR
 */
#define TCONST  1
#define TMEM    2
#define TREGMEM 3
#define TSEG    4
#define TSPCR   5
#define TSPDR   6
#define TSPTR   7
#define TSTRING 8
#define TMATH   9 /* ST(i) math register */

#define S8       1
#define S16      2
#define S32      3
#define SUNKNOWN 4
#define S64      5  /* note: the rest of these are used only with operands */
#define S80      6
#define S48      7
#undef SSHORT /* get rid of def'n from pseudo.h, it's not needed */
#define SSHORT   8 /* jmp sizes:(short only used in block) */
#define SNEAR    9
#define SFAR     10
#define S16o32   11 /* used by constants */

#define RAX     0
#define RCX     1
#define RDX     2
#define RBX     3
#define RSP     4
#define RBP     5
#define RSI     6
#define RDI     7
#define RAH     4
#define RCH     5
#define RDH     6
#define RBH     7

#define RES     0
#define RCS     1
#define RSS     2
#define RDS     3
#define RFS     4
#define RGS     5

FILE *ifile;
FILE *outfile;
int icnt/*instruction count of struct*/;
int mode,lastinstrend,lastgetstrret;
int numberbase=16;

/*MERE: use lastgetstrret always instead of lastinstrend?*/
#define MD386 1 /* modes */
#define MD86  2
#define MDcom 3
#define RS86  1 /* instruction set restrictions */
#define RS286 2 /*HEREfuture: not really implemented*/
#define RS386 3
#define PCBLOCKS 0 /* blocks is normal */
#define PCEXTERN 2 /* used for init. extern vars */

int outplace=PCBLOCKS;
int typegotsize;/* flag indicating last typeget also got size (if==1)*/
int typesize;   /* the size that typeget obtained */
int expseg,expsz,defsz,prst,jccnear,fatal;
int inline;
char *infile,errstr[STRINGSZ];

/* modes:
 * expseg - must explicitely specify segment or warning issued
 * expsz  - must explicitely specify address&opperand size prefixes or warn
 * defsz  - default address&opperand size
 * prst   - processor restrictions on instruction set
 * jccnear- jcc near instruction legal?
 */
main(argcnt,argv) char *argv[];
{ char temp[STRINGSZ],instr[STRINGSZ];
  int v,low,high,mid,eofyet,fcnt,b,a;
  struct varxref *vx;
  struct variables *svs;
  infile="no file";
  for(icnt=0;iset[icnt].func;icnt++)
  { if(icnt&&strcmp(iset[icnt-1].nm,iset[icnt].nm)>=0)
    { sprintf(temp,"table error1: \"%s\" out of alphabetical order"
        ,iset[icnt].nm);
      fatalerror(temp);
    }
  }
  fcnt=2; /* unless all the folling fail, we start at 2 */
  if(!strcmp(argv[1],"386"))
  { mode=MD386;
    defsz=S32;
  }
  else if(!strcmp(argv[1],"86"))
  { mode=MD86;
    defsz=S16;
  }
  else
  { if(strcmp(argv[1],"com")) fcnt=1;
    mode=MDcom;
    expseg=0; expsz=0; defsz=S16; prst=RS386; jccnear=1;
  }
if(mode!=MDcom&&mode!=MD386)
 fatalerror("mode com or 386 is the only one implemented presently");
  AL(blk,1,struct outblock);
  AL(vxref,1,struct varxref);
  AL(svars,1,struct variables);
  AL(stkvars,1,struct stackvars);
  AL(exref,1,struct externxref);
  AL(exttbl,1,struct exttbl);
  AL(extxref,1,struct prexref);
  for(;fcnt<argcnt;fcnt++)
  { infile=argv[fcnt];
    ifile=fopen(infile,"r");
    if(!ifile) { fatalerror("cannot open file"); exit(-1); }
    inline=1;
    strcpy(temp,argv[fcnt]);
    b=strlen(temp);
    if(b<3||temp[b-2]!='.'||(temp[b-1]!='a'&&temp[b-1]!='i'))
      fatalerror("illegal ending");
    switch(mode)
    { case MDcom:
       strcpy(temp+(b-1),"com");
      break;
      case MD386:
       strcpy(temp+(b-1),"o");
      break;
      default: fatalerror("illegal mode:exitting"); exit(-1);
    }
    outfile=fopen(temp,"w");
    if(!outfile) { fatalerror("can not open output file"); exit(-2); }
    stkoff=stkptrexist=stkvcnt=blkcnt=vxcnt=varcnt=0;
    extxcnt=exttblsz=exrcnt=0;
    stkvAL=BUFSZ;
    RAL(exref,1,struct externxref);
    RAL(blk,1,struct outblock);
    RAL(vxref,1,struct varxref);
    RAL(svars,1,struct variables);
    RAL(stkvars,BUFSZ,struct stackvars);
    RAL(exttbl,1,struct exttbl);
    RAL(extxref,1,struct prexref);
    if(mode==MDcom) blk->inoff=0x100;
    else blk->inoff=0;
    AL(blk->bops,1,char);
    blk->blen=0;
    lastgetstrret='\n';
    for(eofyet='A'/*just make sure not EOF*/;eofyet!=EOF;)
    { eofyet=getinstr(temp);
      if(!*temp)
      { if(eofyet!=EOF) fatalerror("confusing situation");
       break;
      }
      low=0; high=icnt-1;
      while(1)
      { if(low>high) break;
        mid=(low+high)/2;
        v=strcmp(iset[mid].nm,temp);
        if(v>0) high=mid-1;
        else if(v<0) low=mid+1;
        else
        { (*iset[mid].func)(iset+mid);
         goto endofsearch;
        }
      }
      a=labeler(temp,strlen(temp));
      if(svars[a].here||(svars[a].tp!=VTABSOLUTE&&svars[a].num!=varcnt-1))
      { sprintf(temp,"multiple definitions of \"%s\" here %d num %d varcnt %d a %d"
,svars[a].nm,svars[a].here,svars[a].num,varcnt,a);
        fatalerror(temp);
      }
      svars[a].here=1;
      svars[a].tp=VTABSOLUTE;
      svars[a].u.bp.blk=blkcnt;
      svars[a].u.bp.off=blk[blkcnt].blen;
     endofsearch: ;
    }
    switch(mode)
    { case MDcom:
       xreffill(0);
       for(a=0;a<=blkcnt;a++)
       { fwrite(blk[a].bops,sizeof(char),blk[a].blen,outfile);
         free(blk[a].bops);
       }
      break;
      case MD386:
       if(blkcnt||blk->blen) fatalerror("end without endfunc");
       for(a=1+2*sizeof(long)+extxcnt
            *sizeof(struct prexref),b=0;b<exttblsz;b++)
        a+=strlen(exttbl[b].nm)+4+4*(exttbl[b].tp==SPOINTER
         ||exttbl[b].tp==SSTRUCT||exttbl[b].tp==SARRAY)
         +exttbl[b].prepres*(sizeof(long)+exttbl[b].presz);
       fwrite(&a,sizeof(long),1,outfile);
       a=OPTEXTERN;
       fwrite(&a,1,1,outfile);
       fwrite(&exttblsz,sizeof(long),1,outfile);
       for(a=0;a<exttblsz;a++)
       { fwrite(exttbl[a].nm,1,strlen(exttbl[a].nm)+1,outfile);
         fwrite(&exttbl[a].scope,1,1,outfile);
         fwrite(&exttbl[a].tp,1,1,outfile);
         if(exttbl[a].tp==SPOINTER||exttbl[a].tp==SSTRUCT
            ||exttbl[a].tp==SARRAY)
           fwrite(&exttbl[a].sz,sizeof(long),1,outfile);
         fwrite(&exttbl[a].prepres,1,1,outfile);
         if(exttbl[a].prepres)
         { fwrite(&exttbl[a].presz,sizeof(long),1,outfile);
           fwrite(exttbl[a].prestf,1,exttbl[a].presz,outfile);
         }
       }
       fwrite(&extxcnt,sizeof(long),1,outfile);
       fwrite(extxref,sizeof(struct prexref),extxcnt,outfile);
       a=0;
       fwrite(&a,4,1,outfile); /* internal external count */
      break;
      default: fatalerror("unknown mode");
    }
    fclose(outfile); fclose(ifile);
    free(vars);
  }
 exit(fatal);
}

xreffill(limvar)
/* if limvar=1 then its mode is such that variables should be stack
 *   or extern., also, variables that aren't local are dumped
 */
{ struct variables *svs;
  char temp[STRINGSZ];
  struct varxref *vx;
  int a,b,again,offset;
  AL(vars,varcnt,varcnttp);
  for(a=0;a<varcnt;a++)
  { vars[svars[a].num]=a;
    if(limvar&&svars[a].tp!=VTLABEL&&svars[a].tp!=VTSTACK
      &&svars[a].tp!=VTEXTERN)
    { sprintf(temp,"var \"%s\" not rel.; off not asm right",svars[a].nm);
      warning(temp);
    }
    if(!svars[a].here&&svars[a].tp!=VTEXTERN)
    { sprintf(temp,"variable \"%s\" not here",svars[a].nm);
      fatalerror(temp);
    }
  }
  do /* fill in blkjmps */
  { struct outblock *bck;
    again=0;
    for(a=0,bck=blk;a<blkcnt;a++,bck++)
    { offset=bck->jloc.off-bck->blen-bck->inoff;
      if(bck->jloc.jloctp==JLLABEL)
      { b=vars[bck->jloc.num];
        if(svars[b].tp!=VTLABEL)
        { sprintf(errstr,"illegal type of jump loc line %d",bck->line);
          fatalerror(errstr);
        }
        offset+=blk[svars[b].u.bp.blk].inoff+svars[b].u.bp.off;
      }
      else if(bck->jloc.jloctp!=JLCONST)
      { inline=bck->line; fatalerror("1unknown jump location type"); }
      if((offset<-128||offset>127)&&bck->jsz!=SNEAR)
      { if(bck->jsz!=SSHORT) fatalerror("internal error blk jmp");
        again=bck->blen;
        /* again used as temporary variable for calculating increase sz */
        makenear(bck->jtp,bck->opsz,a);
        again=bck->blen-again;
        for(b=a+1;b<=blkcnt;b++) blk[b].inoff+=again;
        again=1; bck->jsz=SNEAR;
       continue; /* no point in putnearing it when another chance */
      }
      if(again) continue;
      if(bck->jsz==SNEAR) putnear(bck->jtp,bck->jop,bck->opsz,a,offset);
      else putshort(bck->jop,a,offset);
    }
  } while(again);
/* xref absvars and labxref */
  for(a=0,vx=vxref;a<vxcnt;a++,vx++)
  { switch(vx->tp)
    { case VTLABEL:
       offset=vx->u.lab.jloc.off;
       if(vx->u.lab.isoff) offset-= 1+blk[vx->blk].inoff+vx->off;
       else if(limvar)
       { inline=vx->line;
         fatalerror("not a relative address");
       }
       if(vx->u.lab.jloc.jloctp==JLLABEL)
       { svs=svars+vars[vx->u.lab.jloc.num];
         if(svs->tp!=vx->tp)
         { sprintf(errstr,"1unknown variable type for \"%s\"",svs->nm);
           fatalerror(errstr);
         }
         offset+=svs->u.bp.off+blk[svs->u.bp.blk].inoff;
         if(!svs->here)
         { inline=vx->line;
           fatalerror("variable not here");
         }
       }
       else
       { if(vx->u.lab.jloc.jloctp!=JLCONST)
           fatalerror("2unknown jump location type");
       }
       switch(vx->u.lab.jsz)
       { case SSHORT:
          if(offset>127||offset<-128)
          { inline=vx->line;
            fatalerror("can't fit jump offset in");
          }
          blk[vx->blk].bops[vx->off]=offset;
         break;
         case SNEAR:
          switch(vx->u.lab.opsz)
          { case S16:
             if(vx->u.lab.isoff) offset--;
             *(wordptr)(blk[vx->blk].bops+vx->off)=offset;
            break;
            case S32:
             if(vx->u.lab.isoff) offset-=3;
             *(dwordptr)(blk[vx->blk].bops+vx->off)=offset;
            break;
            default:
             inline=vx->line;
             fatalerror("unknown operand size when filling xrefs");
          }
         break;
         default:
          inline=vx->line;
          fatalerror("unknown jump size when filling xrefs");
       }
      break;
      case VTABSOLUTE:
       svs=svars+vars[vx->vnum];
       if(svs->tp!=vx->tp)
       { inline=vx->line;
         fatalerror("variable type mismatch");
       }
       offset=svs->u.bp.off+blk[svs->u.bp.blk].inoff;
       if(!svs->here)
       { inline=vx->line;
         fatalerror("variable not here");
       }
       switch(vx->u.abs.addsz)
       { case S16:
          *(wordptr)(blk[vx->blk].bops+vx->off)+=offset;
         break;
         case S32:
          *(dwordptr)(blk[vx->blk].bops+vx->off)+=offset;
         break;
         default:
          inline=vx->line; fatalerror("unknwon address size");
       }
      break;
      case VTEXTERN:
       RAL(exref,exrcnt+1,struct externxref);
       exref[exrcnt].off=blk[vx->blk].inoff+vx->off;
       svs=svars+vars[vx->vnum];
       exref[exrcnt].scoff=svs->u.off;
       exref[exrcnt++].ptrfnc=vx->u.ext.ptrfnc;
if(vx->u.ext.addsz!=S32) warning("bad size VTEXTERN");
      break;
      default: fatalerror("unknown var type in vxref");
    }
  }
}

addfill(m) struct memindx m;
/* add m.stf to output  and generate xrefs
 * (prefixes and opcodes not handled here) */
{ int a;
  outoff off;
  off=blk[blkcnt].blen+m.xoff;
  for(a=0;a<m.sz;a++) AB(m.stf[a]);
  if(m.xexist)
  { RAL(vxref,vxcnt+1,struct varxref);
    vxref[vxcnt].blk=blkcnt;
    vxref[vxcnt].off=off;
    vxref[vxcnt].line=inline;
    vxref[vxcnt].tp=m.xexist; /* no constant labels will appear in addfill*/
    switch(m.xexist)
    { case VTABSOLUTE:
       vxref[vxcnt].u.abs.addsz=m.addsz;
       vxref[vxcnt].vnum=svars[m.scoff].num;
      break;
      case VTLABEL:
       vxref[vxcnt].u.lab.jloc=m.jl; /* copies vnum here */
       vxref[vxcnt].u.lab.opsz=m.addsz;
       vxref[vxcnt].u.lab.isoff=0;
       vxref[vxcnt].u.lab.jsz=SNEAR;
      break;
      case VTEXTERN:
       vxref[vxcnt].vnum=svars[m.scoff].num;
       if(exttbl[svars[m.scoff].u.off].tp==SFUNC) vxref[vxcnt].u.ext.ptrfnc=1;
       else vxref[vxcnt].u.ext.ptrfnc=0;
       vxref[vxcnt].u.ext.addsz=m.addsz;
      break;
      default:
       fatalerror("unknown variable type in addfill");
    }
    vxcnt++;
  }
}

prefix(repok,seg,opsz,addsz)
{ if(pfx.ipre)
  { if(!repok&&pfx.ipre!=0xf0) warning("strong: rep at illegal spot");
    AB(pfx.ipre);
  }
  if(addsz||pfx.addsz)
  { if((addsz!=pfx.addsz&&expsz)||pfx.addsz&&!addsz)
      warning("address size prefix inconsistency");
    AB(0x67);
  }
  if(opsz||pfx.opsz)
  { if((opsz!=pfx.opsz&&expsz)||pfx.opsz&&!opsz)
      warning("opperand size prefix inconsistency");
    AB(0x66);
  }
  if(seg||pfx.seg)
  { if(seg!=pfx.seg&&(expseg||(pfx.seg&&seg)))/*HEREfutrure assume*/
      warning("segment prefix inconsistency");
    if(seg) AB(seg);
    else AB(pfx.seg);
  }
}

getinstr(s) char *s; /* get prefixes(in struct) and instruction(into string) */
{ char temp[STRINGSZ];
  int a,b,c,isprefix=1,oldlastgetstrret;
  pfx.ipre=pfx.seg=pfx.opsz=pfx.addsz=0;
  do
  { /* get string */
    oldlastgetstrret=lastgetstrret;
    lastinstrend=c=getstr(temp,' ',1);
    if(isprefix==1&&oldlastgetstrret!='\n'&&strcmp(temp,"db")
       &&strcmp(temp,"dw")&&strcmp(temp,"ddw"))
      warning("get instr. which is not 1st string on line");
    /* cmp with prefixlist  in optimized manner*/
    if(c==':')
    { switch(*temp)
      { case 'c':/*cs*/
         if(strcmp(temp,"cs")) goto islabel;
         if(pfx.seg) warning("multiple segment prefixes, using last");
         pfx.seg=0x2e;
        break;
        case 's':/*ss*/
         if(strcmp(temp,"ss")) goto islabel;
         if(pfx.seg) warning("multiple segment prefixes, using last");
         pfx.seg=0x36;
        break;
        case 'd':/*ds*/
         if(strcmp(temp,"ds")) goto islabel;
         if(pfx.seg) warning("multiple segment prefixes, using last");
         pfx.seg=0x3e;
        break;
        case 'e':/*es*/
         if(strcmp(temp,"es")) goto islabel;
         if(pfx.seg) warning("multiple segment prefixes, using last");
         pfx.seg=0x26;
        break;
        case 'f':/*fs*/
         if(strcmp(temp,"fs")) goto islabel;
         if(pfx.seg) warning("multiple segment prefixes, using last");
         pfx.seg=0x64;
        break;
        case 'g':/*gs*/
         if(strcmp(temp,"gs")) goto islabel;
         if(pfx.seg) warning("multiple segment prefixes, using last");
         pfx.seg=0x65;
        break;
        case 'o':/*opperandsize*/
         if(strcmp(temp,"opperandsize")) goto islabel;
         if(pfx.opsz) warning("unnecessary many opparandsize prefixes");
         pfx.opsz=1;
        break;
        case 'a':/*addresssize*/
         if(strcmp(temp,"addresssize")) goto islabel;
         if(pfx.addsz) warning("unnecessary many addresssize prefixes");
         pfx.addsz=1;
        break;
        default:
        islabel: /* LABEL */
         if(pfx.seg||pfx.opsz||pfx.addsz||pfx.ipre)
           warning("label between prefix and opcode is moved to before prefix");
         a=labeler(temp,strlen(temp));
         if(svars[a].here||svars[a].tp!=VTLABEL)
         { fatalerror("multiple labels");
fprintf(stderr,"here %d a %d tp %d\n",svars[a].here,a,svars[a].tp);
}
         svars[a].tp=VTLABEL;
         svars[a].u.bp.blk=blkcnt;
         svars[a].u.bp.off=blk[blkcnt].blen;
         svars[a].here=1;
      }
    }
    else
    { if(!strcmp(temp,"lock"))
      { if(pfx.ipre) warning("multiple instruction prefixes");
        pfx.ipre=0xf0;
      }
      else if(*temp=='r'&&temp[1]=='e')
      { if(!strcmp(temp,"rep"))
        { if(pfx.ipre) warning("multiple instruction prefixes");
          pfx.ipre=0xf3;
        }
        else if(!strcmp(temp,"repe"))
        { if(pfx.ipre) warning("multiple instruction prefixes");
          pfx.ipre=0xf3;
        }
        else if(!strcmp(temp,"repz"))
        { if(pfx.ipre) warning("multiple instruction prefixes");
          pfx.ipre=0xf3;
        }
        else if(!strcmp(temp,"repne"))
        { if(pfx.ipre) warning("multiple instruction prefixes");
          pfx.ipre=0xf2;
        }
        else if(!strcmp(temp,"repnz"))
        { if(pfx.ipre) warning("multiple instruction prefixes");
          pfx.ipre=0xf2;
        }
        else isprefix=0;
      }
      else isprefix=0;
    }
    if(isprefix) isprefix=2; /* flag that second time through loop */
  } while(isprefix&&c!=EOF);
  /* must be instruction/ return it */
  if(c==EOF&&isprefix) fatalerror("unexpected EOF with i. prefix only");
  strcpy(s,temp);
 return c;
}

int openbracket,closebracket; /* used with address operand calculation */
struct operand getoperand(seperator)
{ char temp[STRINGSZ];
#define MAXOPSZ 8
  static struct operandregs
  { char op[MAXOPSZ];
    struct operand out;
  } opset[]=
  { {"ah" ,{0,TREGMEM,RAH,S8 ,SUNKNOWN,0,0,0,0,0,{{0xc4},1,VTNONE}}},
    {"al" ,{0,TREGMEM,RAX,S8 ,SUNKNOWN,0,0,0,0,0,{{0xc0},1,VTNONE}}},
    {"ax" ,{0,TREGMEM,RAX,S16,SUNKNOWN,0,0,0,0,0,{{0xc0},1,VTNONE}}},
    {"bh" ,{0,TREGMEM,RBH,S8 ,SUNKNOWN,0,0,0,0,0,{{0xc7},1,VTNONE}}},
    {"bl" ,{0,TREGMEM,RBX,S8 ,SUNKNOWN,0,0,0,0,0,{{0xc3},1,VTNONE}}},
    {"bp" ,{0,TREGMEM,RBP,S16,SUNKNOWN,0,0,0,0,0,{{0xc5},1,VTNONE}}},
    {"bx" ,{0,TREGMEM,RBX,S16,SUNKNOWN,0,0,0,0,0,{{0xc3},1,VTNONE}}},
    {"ch" ,{0,TREGMEM,RCH,S8 ,SUNKNOWN,0,0,0,0,0,{{0xc5},1,VTNONE}}},
    {"cl" ,{0,TREGMEM,RCX,S8 ,SUNKNOWN,0,0,0,0,0,{{0xc1},1,VTNONE}}},
    {"cr0",{0,TSPCR  ,0  ,S32,SUNKNOWN,0,0,0,0,0}},
    {"cr2",{0,TSPCR  ,2  ,S32,SUNKNOWN,0,0,0,0,0}},
    {"cr3",{0,TSPCR  ,3  ,S32,SUNKNOWN,0,0,0,0,0}},
    {"cs" ,{0,TSEG   ,RCS,S16,SUNKNOWN,0,0,0,0,0}},
    {"cx" ,{0,TREGMEM,RCX,S16,SUNKNOWN,0,0,0,0,0,{{0xc1},1,VTNONE}}},
    {"dh" ,{0,TREGMEM,RDH,S8 ,SUNKNOWN,0,0,0,0,0,{{0xc6},1,VTNONE}}},
    {"di" ,{0,TREGMEM,RDI,S16,SUNKNOWN,0,0,0,0,0,{{0xc7},1,VTNONE}}},
    {"dl" ,{0,TREGMEM,RDX,S8 ,SUNKNOWN,0,0,0,0,0,{{0xc2},1,VTNONE}}},
    {"dr0",{0,TSPDR  ,0  ,S32,SUNKNOWN,0,0,0,0,0}},
    {"dr1",{0,TSPDR  ,1  ,S32,SUNKNOWN,0,0,0,0,0}},
    {"dr2",{0,TSPDR  ,2  ,S32,SUNKNOWN,0,0,0,0,0}},
    {"dr3",{0,TSPDR  ,3  ,S32,SUNKNOWN,0,0,0,0,0}},
    {"dr6",{0,TSPDR  ,6  ,S32,SUNKNOWN,0,0,0,0,0}},
    {"dr7",{0,TSPDR  ,7  ,S32,SUNKNOWN,0,0,0,0,0}},
    {"ds" ,{0,TSEG   ,RDS,S16,SUNKNOWN,0,0,0,0,0}},
    {"dx" ,{0,TREGMEM,RDX,S16,SUNKNOWN,0,0,0,0,0,{{0xc2},1,VTNONE}}},
    {"eax",{0,TREGMEM,RAX,S32,SUNKNOWN,0,0,0,0,0,{{0xc0},1,VTNONE}}},
    {"ebp",{0,TREGMEM,RBP,S32,SUNKNOWN,0,0,0,0,0,{{0xc5},1,VTNONE}}},
    {"ebx",{0,TREGMEM,RBX,S32,SUNKNOWN,0,0,0,0,0,{{0xc3},1,VTNONE}}},
    {"ecx",{0,TREGMEM,RCX,S32,SUNKNOWN,0,0,0,0,0,{{0xc1},1,VTNONE}}},
    {"edi",{0,TREGMEM,RDI,S32,SUNKNOWN,0,0,0,0,0,{{0xc7},1,VTNONE}}},
    {"edx",{0,TREGMEM,RDX,S32,SUNKNOWN,0,0,0,0,0,{{0xc2},1,VTNONE}}},
    {"es" ,{0,TSEG   ,RES,S16,SUNKNOWN,0,0,0,0,0}},
    {"esi",{0,TREGMEM,RSI,S32,SUNKNOWN,0,0,0,0,0,{{0xc6},1,VTNONE}}},
    {"esp",{0,TREGMEM,RSP,S32,SUNKNOWN,0,0,0,0,0,{{0xc4},1,VTNONE}}},
    {"fs" ,{0,TSEG   ,RFS,S16,SUNKNOWN,0,0,0,0,0}},
    {"gs" ,{0,TSEG   ,RGS,S16,SUNKNOWN,0,0,0,0,0}},
    {"si" ,{0,TREGMEM,RSI,S16,SUNKNOWN,0,0,0,0,0,{{0xc6},1,VTNONE}}},
    {"sp" ,{0,TREGMEM,RSP,S16,SUNKNOWN,0,0,0,0,0,{{0xc4},1,VTNONE}}},
    {"ss" ,{0,TSEG   ,RSS,S16,SUNKNOWN,0,0,0,0,0}},
    {"st" ,{0,TMATH  ,0  }},
    {"st(0)",{0,TMATH,0  }},
    {"st(1)",{0,TMATH,1  }},
    {"st(2)",{0,TMATH,2  }},
    {"st(3)",{0,TMATH,3  }},
    {"st(4)",{0,TMATH,4  }},
    {"st(5)",{0,TMATH,5  }},
    {"st(6)",{0,TMATH,6  }},
    {"st(7)",{0,TMATH,7  }},
    {"tr6",{0,TSPTR  ,6  ,S32,SUNKNOWN,0,0,0,0,0}},
    {"tr7",{0,TSPTR  ,7  ,S32,SUNKNOWN,0,0,0,0,0}},
    {""}
  };
  struct operand out;
  static int oplen=0;
  int v,c,mid,low,high;
  int a,base,base2,indx,usefulloff,mult,offset,off,inbp,islabel;
  int basic,sib,scale,seggood,c2,len,moff,reg,isbp;
  /* off is off of temp; offset is const off from base for mem addrs*/
  if(!oplen) /* if haven't yet counted operands */
  { for(;*opset[oplen].op;oplen++)
    { if(oplen&&strcmp(opset[oplen-1].op,opset[oplen].op)>=0)
      { sprintf(temp,"table error2: \"%s\" # %d out of alphabetical order"
          ,opset[oplen].op,oplen);
        fatalerror(temp);
      }
    }
    if(!oplen) fatalerror("no operands in table");
  }
  if(lastinstrend=='\n') fatalerror("attempting to get operand on new line");
  c=getstr(temp,seperator,0);
 /* test for constant */
  if(*temp=='"')
  { strcpy(out.string,temp+1); /* erase first character */
    out.tp=TSTRING;
    out.lastop=c=='\n';
   return out;
  }

   /* 2 lines- make sure constants begin with digit, find end of const */
  if(isdigit(*temp)) a=find_non_number(temp);
  else a=0;
/*for(;isdigit(temp[a])||(temp[a]<='f'&&temp[a]>='a');a++);*/
  if(!temp[a]) /* true if constant*/
  { out.tp=TCONST;
    out.sign=c;
    out.cval=getnum(temp,strlen(temp));
    if(c=='-') out.cval= -out.cval;
    if(out.cval<128&&out.cval>=-128) out.opsz=S8;
    else out.opsz=S16o32;
    out.lastop=c=='\n';
   return out;
  }
  if(*temp=='\'')
   /* true if char constant*/
  { out.tp=TCONST;
    out.sign=c;
    out.cval=temp[1];
    if(c=='-') warning("ignoring '-' 'char const'"),out.cval= -out.cval;
    if(out.cval<128&&out.cval>=-128) out.opsz=S8;
    else out.opsz=S16o32;
    out.lastop=c=='\n';
   return out;
  }
  low=0; high=oplen-1;
  /* scan for it in opset */
  while(1) /* binary search for regs */
  { if(low>high) break;
    mid=(low+high)/2;
    v=strcmp(opset[mid].op,temp);
    if(v>0) high=mid-1;
    else if(v<0) low=mid+1;
    else
    { out=opset[mid].out;
      out.lastop=c=='\n';
     return out;
    }
  }
  /* if not there, must be memory/ work on it */
  out.tp=TMEM; out.lastop=c=='\n';
  out.addsz=out.jsz=out.opsz=SUNKNOWN;
  if((!strncmp(temp,"word ptr",8)&&!isalnum(temp[8]))
    ||(!strncmp(temp,"wo",2)&&!isalnum(temp[2])))
  { if(!isalnum(temp[2])) off=2;
    else off=8;
    out.opsz=S16;
  }
  else if((!strncmp(temp,"byte ptr",8)&&!isalnum(temp[8]))
    ||(!strncmp(temp,"by",2)&&!isalnum(temp[2])))
  { if(!isalnum(temp[2])) off=2;
    else off=8;
    out.opsz=S8;
  }
  else if((!strncmp(temp,"dword ptr",9)&&!isalnum(temp[9]))
    ||(!strncmp(temp,"dw",2)&&!isalnum(temp[2])))
  { if(!isalnum(temp[2])) off=2;
    else off=9;
    out.opsz=S32;
  }
  else if((!strncmp(temp,"qword ptr",9)&&!isalnum(temp[9]))
    ||(!strncmp(temp,"qw",2)&&!isalnum(temp[2])))
  { if(!isalnum(temp[2])) off=2;
    else off=9;
    out.opsz=S64;
  }
  else if((!strncmp(temp,"tenbyte ptr",11)&&!isalnum(temp[11]))
    ||(!strncmp(temp,"tb",2)&&!isalnum(temp[2])))
  { if(!isalnum(temp[2])) off=2;
    else off=11;
    out.opsz=S80;
  }
  else if((!strncmp(temp,"pword ptr",9)&&!isalnum(temp[9]))
    ||(!strncmp(temp,"pw",2)&&!isalnum(temp[2])))
  { if(!isalnum(temp[2])) off=2;
    else off=9;
    out.opsz=S48;
  }
  else if((!strncmp(temp,"near",4)&&!isalnum(temp[4]))
    ||(!strncmp(temp,"ne",2)&&!isalnum(temp[2])))
  { if(!isalnum(temp[2])) off=2;
    else off=4;
    out.jsz=SNEAR;
  }
  else if(!strncmp(temp,"far",3)&&(!isalnum(temp[3])))
  { off=3;
    out.jsz=SFAR;
  }
  else
  { out.opsz=SUNKNOWN;
    off=0;
  }
  base=base2=indx= -1;
  openbracket=closebracket=usefulloff=mult=offset=0;
  out.m.xexist=VTNONE; out.tplab=seggood=islabel=0;
  out.m.jl.jloctp=JLCONST;
  for(;temp[off]&&(temp[off]!=']'||temp[off+1]);off+=len)
  { c2=getpiece(temp,&off,&len);
    if(c2!='['&&c2!=']'&&c2!='\0'&&c2!='+'&&c2!='-'&&c2!=':'&&c2!='*')
{
fprintf(stderr,"\"%c\" [%s]\n",c2,temp);
       fatalerror("illegal seperator");
}
    if(isdigit(temp[off])) /*CONSTANT*/
    { a=getnum(temp+off,len);
      if(c2=='-') a= -a;
      if(mult) { a*=moff; mult=0; }
      if(temp[off+len]=='*') { mult=1; moff=a; }
      else if(temp[off+len]==':')
      { if(mult||offset||seggood)
           fatalerror("only non-complex segs supported");
        out.seg=a;
        if(out.jsz!=SFAR&&out.jsz!=SUNKNOWN)
          fatalerror("far jump already declared otherwise");
        out.jsz=SFAR;
        openbracket++; closebracket++;
        /* don't trigger bracket error when both constant */
        seggood=islabel=1;
      }
      else offset+=a;
    }
    else if(len==3&&temp[off]=='e') /* possibly 32 bit register */
    { reg=regnum(temp+(off+1));
      if(reg==-1) goto isname;
      if(c2=='-') fatalerror("can't subtract non-constant");
      if(temp[off+len]=='*')
      { off++; /* acknowledge * */
        if(mult||indx!=-1) fatalerror("multiple scaled indexes");
        indx=reg; scale=powof2(atoi(temp+(off+len)));
        if(scale==-1||!isdigit(temp[off+len])||isdigit(temp[off+len+1]))
          fatalerror("illegal scale for index");
        off++; /* acknowledge digit */
      }
      else if(mult)
      { if(indx!=-1) fatalerror("multiple scaled indexes");
        indx=reg; scale=powof2(moff);
        mult=0;
      }
      else if(base!=-1)
      { if(indx!=-1) fatalerror("too many registers in memory index");
        indx=reg; scale=0;
      }
      else base=reg;
      if(out.addsz!=SUNKNOWN&&out.addsz!=S32)
        fatalerror("32 and 16 bit pieces");
      out.addsz=S32;
    }
    else if(len==2)
    { reg=regnum(temp+off);
      if(reg==-1) goto isname;
      if(base==-1) base=reg;
      else
      { if(base2==-1) fatalerror("too many regs");
        base2=reg;
      }
      if(c2=='-'||c2=='*') fatalerror("can't subtract non-constant");
      if(out.addsz!=SUNKNOWN&&out.addsz!=S16)
        fatalerror("32 and 16 bit pieces");
      out.addsz=S16;
    }
    else
    {isname:
      /* right now we only recognise mode COMfile and absolute variables */
      a=labeler(temp+off,len);
      out.m.scoff=a;
      if(out.m.xexist) fatalerror("multiple variables");
      if(mult) fatalerror("illegal * factor");
      openbracket++; closebracket++;
      switch(svars[a].tp) /*keyword for rapid find: GETVAR*/
      { case VTABSOLUTE:
         usefulloff=1;
         out.m.xexist=VTABSOLUTE;
        break;
        case VTLABEL:
         usefulloff=1;
         out.m.xexist=VTLABEL;
         out.tplab=1;
         out.m.jl.num=svars[a].num;
         out.m.jl.jloctp=JLLABEL;
        break;
        case VTSTACK:
         if(mode!=MD386)
           fatalerror("stack variables not implemented outside 386");
         offset+=svars[a].u.off;
         if(base!=-1)
         { if(defsz==S16)
           { if(base2!=-1) fatalerror("too many registers (stack variable)");
             if(out.addsz!=SUNKNOWN&&out.addsz!=S16)
               fatalerror("address size mismatch");
             base2=RBP;
           }
           else
           { if(defsz!=S32) fatalerror("unknown default size");
             if(indx!=-1)
               fatalerror("too many registers stack variable access");
             if(out.addsz!=SUNKNOWN&&out.addsz!=S32)
               fatalerror("address size mismatch");
             indx=RBP;
             scale=0;
           }
         }
         else base=RBP;
        break;
        case VTEXTERN:
         usefulloff=1;
         out.m.xexist=VTEXTERN;
         if(exttbl[svars[a].u.off].tp==SFUNC)
         { out.tplab=1;
           out.m.jl.num=svars[a].num;
           out.m.jl.jloctp=JLFUNC;
           if(out.jsz!=SNEAR&&out.jsz!=SUNKNOWN)
             fatalerror("illegal jump size with extern");
           else out.jsz=SNEAR;
         }
        break;
        default: fatalerror("3unknown variable type");
      }
    }
  }
  if(islabel) out.tplab=1; /* used with segs */
  if(temp[off]==']') closebracket++;
  if(openbracket!=closebracket||!openbracket)
    fatalerror("bad brackets on address");
  out.m.sz=1;
  out.m.jl.off=offset;
  if(out.addsz==SUNKNOWN) out.addsz=pfx.addsz?(defsz==S32?S16:S32):defsz;
  if(out.addsz==S32)
  { isbp=0; sib= -1;
    if(indx!=-1) /* then SIB present */
    { basic=0x04; /* SIB present indicator */
      if(base==RBP) isbp=1;
      if(base==-1) { base=RBP; usefulloff=1; basic-=0x80;/*counteragent*/ }
      sib=base+(indx<<3)+(scale<<6);
      if(indx==RSP)
      { if(scale) fatalerror("scaled register sp");
        sib=indx+(base<<3);
      }
    }
    else if(base==RSP) /* special needs SIB to address SP */
    { basic=0x04; sib=0x24; }
    else /* no SIB needed */
    { if(base==RBP) isbp=1;
      basic=base;
    }
    if(basic==-1)
    { *out.m.stf=0x05;
      out.m.xoff=out.m.sz;
      *(dwordptr)(out.m.stf+out.m.sz)=offset; out.m.sz+=4;
    }
    else if(!offset&&!isbp&&!usefulloff)
    { *out.m.stf=basic;
      if(sib!=-1) out.m.stf[out.m.sz++]=sib;
    }
    else if(offset<128&&offset>=-128&&!usefulloff)
    { *out.m.stf=basic+0x40;
      if(sib!=-1) out.m.stf[out.m.sz++]=sib;
      out.m.stf[out.m.sz++]=offset;
    }
    else
    { *out.m.stf=basic+0x80;
      if(sib!=-1) out.m.stf[out.m.sz++]=sib;
      out.m.xoff=out.m.sz;
      *(dwordptr)(out.m.stf+out.m.sz)=offset; out.m.sz+=4;
    }
  }
  else
  { if(out.addsz!=S16) fatalerror("address size illegal");
    isbp=0;
    switch(base2)
    { case RBP:
       if(base==RSI) basic=0x02;      /*[bp+si]*/
       else if(base==RDI) basic=0x03; /*[bp+di]*/
       else fatalerror("illegal register with 2");
      break;
      case RBX:
       if(base==RSI) basic=0x00;      /*[bp+si]*/
       else if(base==RDI) basic=0x01; /*[bp+di]*/
       else fatalerror("illegal register with 2");
      break;
      case RSI:
       if(base==RBP) basic=0x02;      /*[bp+si]*/
       else if(base==RBX) basic=0x00; /*[bp+di]*/
       else fatalerror("illegal register with 2");
      break;
      case RDI:
       if(base==RBP) basic=0x03;      /*[bp+si]*/
       else if(base==RBX) basic=0x01; /*[bp+di]*/
       else fatalerror("illegal register with 2");
      break;
      case -1: /* one or fewer regs used in address */
       switch(base)
       { case RBP:
          basic=0x06;
          isbp=1;
         break;
         case RBX:
          basic=0x07;
         break;
         case RSI:
          basic=0x04;
         break;
         case RDI:
          basic=0x05;
         break;
         case -1: /* no registers used in address; just offset */
          basic= -1;
         break;
         default: fatalerror("internal base");
       }
      break;
      default: fatalerror("internal base2");
    }
    if(basic==-1)
    { *out.m.stf=0x06;
      out.m.xoff=out.m.sz;
      *(wordptr)(out.m.stf+out.m.sz)=offset; out.m.sz+=2;
    }
    else if(!offset&&!isbp&&!usefulloff)
    { *out.m.stf=basic;
    }
    else if(offset<128&&offset>=-128&&!usefulloff)
    { *out.m.stf=basic+0x40;
      out.m.stf[out.m.sz++]=offset;
    }
    else
    { *out.m.stf=basic+0x80;
      out.m.xoff=out.m.sz;
      *(wordptr)(out.m.stf+out.m.sz)=offset; out.m.sz+=2;
    }
  }
  out.m.addsz=out.addsz;
 return out;
}

varcnttp findvar(s,l) char *s; int l;
/* returns index into slab of string s of length l
 *  or returns -1-(position where it should be) */
{ int c,v;
  varcnttp mid,low,high;
  c=s[l];
  s[l]='\0';
  low=0; high=varcnt-1;
  while(1)
  { if(low>high) break;
    mid=(low+high)/2;
    v=strcmp(svars[mid].nm,s);
    if(v<0) low=mid+1;
    else if(v>0) high=mid-1;
    else /*v==0*/
    { s[l]=c;
     return mid;
    }
  } /* not in table, insert it at low */
  s[l]=c;
 return -1-low;
}

varcnttp labeler(s,l) char *s; int l;
{ varcnttp vn;
  struct variables *svs;
  vn=findvar(s,l);
  if(vn>=0) return vn;
  vn= -1-vn; /* get position where it should be */
  if(lastextpres&&vn<lastextnum) lastextnum++;
  RAL(svars,varcnt+1,struct variables);
  svs=svars+vn;
  memmove(svs+1,svs,(varcnt-vn)*sizeof(struct variables));
  svs->here=0;
  svs->local=1;
  svs->tp=VTLABEL;
  svs->u.bp.blk=blkcnt;
  svs->u.bp.off=blk[blkcnt].blen;
  svs->num=varcnt++;
  AL(svs->nm,l+1,char);
  memcpy(svs->nm,s,l);
  svs->nm[l]='\0';
 return vn;
}

getpiece(s,off,len) char *s; int *off,*len;
/* get one piece(item) of an address return seperator at BEGINNING
 * update open&close bracket for housekeeping*/
{ int dash=0,ret='\0',prev='\0';
  for(;s[*off]&&!isalnum(s[*off])&&s[*off]!='_';(*off)++) /*check seperators */
  { ret=s[*off];
    if(ret=='-')
    { if(prev&&prev!=']') fatalerror("messed up seperators 1");
      dash=1;
    }
    else if(ret=='[')
    { if(prev&&prev!='+'&&prev!='-') fatalerror("messed up seperators 2");
      openbracket++;
    }
    else if(ret==']')
    { closebracket++;
      if(prev||closebracket>openbracket) fatalerror("messed up seperators 3");
    }
  }
  for(*len=0;isalnum(s[*off+*len])||s[*off+*len]=='_';(*len)++); /*read main */
  if(dash) ret='-';
  if(ret==' ') ret=0;
 return ret;
}

regnum(s) char *s;
/* returns reg num RAX... of 16-bit general register of first
 *  2 positions in string s; ignores everything but first 2 chars */
{ int reg= -1;
  switch(*s)
  { case 'a':
     if(s[1]=='x') reg=RAX;
    break;
    case 'b':
     if(s[1]=='x') reg=RBX;
     else if(s[1]=='p') reg=RBP;
    break;
    case 'c':
     if(s[1]=='x') reg=RCX;
    break;
    case 'd':
     if(s[1]=='x') reg=RDX;
     else if(s[1]=='i') reg=RDI;
    break;
    case 's':
     if(s[1]=='i') reg=RSI;
     else if(s[1]=='p') reg=RSP;
  } /*default: reg already==-1*/
 return reg;
}

process_C_comment()
{ int ch,lastch;
  char comment[STRINGSZ+1];
  static char filename[STRINGSZ+1];
  int len,checklen,newline;

  ch=getic();
  if(ch!='*')
  { ungetic(ch);
    ch='/';
  }
  else
  { ch=' '; /* anything but '*' */
    len=0;
    do
    { lastch=ch;
      ch=getic(ch);
      comment[len++]=ch;
    } while(ch!=EOF&&(ch!='/'||lastch!='*')&&len<STRINGSZ);
    if(len>=STRINGSZ)
    { fatalerror("comment length string overflow");
      len=STRINGSZ-1;
    }
    if(ch!=EOF) ch=getic();
    else len--;
    comment[len]='\0';
  }
  len= -1;
  if(sscanf(comment,"%dL%n",&checklen,&len)!=1||len<2
     ||len+checklen+3>strlen(comment)
     ||sscanf(comment+len+checklen,"%d",&newline)!=1||newline<0)
    return ch;
  memcpy(filename,comment+len,checklen); filename[checklen]='\0';
  infile=filename;
  inline=newline+1;
 return ch;
}

getstr(s,end,iflag) char *s,end; /* get stuff until char end, EOL, or EOF */
/* if end==' ' stop on '\t' as well
 * if iflag stop on ':'; if not; no more args return empty
 * return char that stopped it */
{ int ch,i=0,eow=0,endch; /* end of word */
  do /* get first real character */
  { do
    { ch=getic();
      if(ch=='\n'&&!iflag) { *s='\0'; return '\n'; }
      else if(ch=='/') ch=process_C_comment();
    } while(ch==' '||ch=='\n'||ch=='\t');
    if(ch==';')
    { do ch=getic(); while(ch!='\n'&&ch!=EOF);
      if(!iflag) { *s='\0'; return '\n'; }
    }
  } while(ch=='\n');
  if(ch==EOF) { *s='\0'; ch=EOF; }
  else if(ch=='"'||ch=='\'') /* get quoted string */
  { if(iflag) fatalerror("internal error iflag");
    s[i++]=endch=ch;
    do
    { ch=getic();
      if(ch=='\\')
       switch(getic())
       { case 'n':  s[i++]='\n'; break;
         case 't':  s[i++]='\t'; break;
         case 'r':  s[i++]='\r'; break;
         case 'f':  s[i++]='\f'; break;
         case 'b':  s[i++]='\b'; break;
         case '\\': s[i++]='\\'; break;
         case '0':  s[i++]='\0'; break;
         default: warning("ignoring weird char const");
       }
      else if(ch!=endch) s[i++]=ch;
    } while(ch!=endch&&ch!=EOF&&i<STRINGSZ);
    if(endch=='\''&&i!=2)
      warning("weird char. const. prob. not compiled right");
    while(ch!=';'&&ch!='\n'&&ch!=EOF&&ch!=end&&ch!=','&&ch!=' ')
    { ch=getic();
      if(ch=='/') ch=process_C_comment();
    }
  }
  else
  { do /* get actual string */
    { if(ch!=' '&&ch!='\t')              /* cut out white space */
      { if(eow&&isalnum(ch)) s[i++]=' '; /* unless it seperates 2 words */
        s[i++]=ch;
        eow=0;
      }
      else if(isalnum(s[i-1])) eow=1; /* end of word */
      else eow=0;
      ch=getic();
      if(ch=='/') ch=process_C_comment();
    } while(ch!=';'&&ch!='\n'&&ch!=','&&ch!=end
       &&(end!=' '||ch!='\t')&&(!iflag||ch!=':')&&ch!=EOF&&i<STRINGSZ);
  }
  if(ch==' '||ch=='\t')
  { do
    { ch=getic();
      if(ch=='/') ch=process_C_comment();
    } while(ch==' '||ch=='\t');
    if(ch!='\n'&&ch!=';'&&ch!=EOF&&(!iflag||ch!=':')&&ch!=',') ungetic(ch);
  }
  if(ch==';')
    do ch=getic();
    while(ch!='\n'&&ch!=EOF);
  if(i>=STRINGSZ) { fatalerror("string overflow"); i=STRINGSZ-1; }
  s[i]='\0';
  lastgetstrret=ch;
 return ch;
}

getic()
{ int c;
  c=getc(ifile);
  if(c=='\n') inline++;
 return c;
}

ungetic(c)
{ if(c=='\n') inline--;
  ungetc(c,ifile);
}

ADW(dw) long dw;
{ int k;
  if(outplace==PCBLOCKS)
  { if(!(k=blk[blkcnt].blen%BUFSZ)||k>BUFSZ-4)
    { RAL(blk[blkcnt].bops,blk[blkcnt].blen+BUFSZ,char); }
    *(dwordptr)(blk[blkcnt].bops+blk[blkcnt].blen)=dw; blk[blkcnt].blen+=4;
  }
  else if(outplace==PCEXTERN)
  { if(!lastextpres||exttbl[lastextnum].presz>=exttbl[lastextnum].sz)
      fatalerror("too much prestf init in extern");
    else
    { *(dwordptr)(exttbl[lastextnum].prestf+exttbl[lastextnum].presz)=dw;
      exttbl[lastextnum].presz+=4;
    }
  }
  else
    fatalerror("double word init of unknown");
}

AW(w) /* add word w to output */
{ int k;
  if(outplace==PCBLOCKS)
  { if(!(k=blk[blkcnt].blen%BUFSZ)||k>BUFSZ-2)
    { RAL(blk[blkcnt].bops,blk[blkcnt].blen+BUFSZ,char); }
    *(wordptr)(blk[blkcnt].bops+blk[blkcnt].blen)=w; blk[blkcnt].blen+=2;
  }
  else if(outplace==PCEXTERN)
  { if(!lastextpres||exttbl[lastextnum].presz>=exttbl[lastextnum].sz)
      fatalerror("too much prestf init in extern");
    else
    { *(wordptr)(exttbl[lastextnum].prestf+exttbl[lastextnum].presz)=w;
      exttbl[lastextnum].presz+=2;
    }
  }
  else
     fatalerror("unknown outplace in AW");
}

AB(b) /* add byte b to output */
{ if(outplace==PCBLOCKS)
  { if(!(blk[blkcnt].blen%BUFSZ))
    { RAL(blk[blkcnt].bops,blk[blkcnt].blen+BUFSZ,char); }
    blk[blkcnt].bops[blk[blkcnt].blen++]=b;
  }
  else if(outplace==PCEXTERN)
  { if(!lastextpres||exttbl[lastextnum].presz>=exttbl[lastextnum].sz)
      fatalerror("too much prestf init in extern");
    else exttbl[lastextnum].prestf[exttbl[lastextnum].presz++]=b;
  }
  else
     fatalerror("unknown output place");
}

makeshort() /* make room for short jump in latest block */
{ AW(0); }

makenear(jtp,opsz,bcnt)
/* make additional room to go from short jump to long jump */
{ blk[bcnt].blen+=(jtp==JTJcc)+1+2*(opsz==S32);
  RAL(blk[bcnt].bops,blk[bcnt].blen+BUFSZ,char);
}

putshort(op,bcnt,off) outoff off;
{ blk[bcnt].bops[blk[bcnt].blen-2]=op;
  blk[bcnt].bops[blk[bcnt].blen-1]=off;
}

putnear(jtp,op,opsz,bcnt,off) outoff off;
{ int sz;
  char *pc;
  pc=blk[bcnt].bops+blk[bcnt].blen;
  if(opsz==S32) sz=4;
  else
  { if(opsz!=S16) fatalerror("can't keep track of opperand size");
    sz=2;
  }
  switch(jtp)
  { case JTJcc:
     if(op<0x80&&op>=0x70) op+=0x10;
     else fatalerror("cannot increase size of jump");
     pc[-2-sz]=0x0f; /* 2 byte code */
    break;
    case JTJMP:
     op=0xe9;
    case JTCALL:
    break;
    default: fatalerror("illegal jump type");
  }
  pc[-1-sz]=op;
  if(opsz==S32) *(dwordptr)(pc-sz)=off;
  else *(wordptr)(pc-sz)=off;
}

int find_non_number(s) char *s;
/* returns the location of the first digit that is inconsistent with being part
 *  of a number as defined by the current mode (HEX/DECIMAL/OCTAL) */
/* PORT: this assumes that 0..9 are in order and nothing is in between in
 *  charset*/
{ int a=0;
  char maxdigit='9'; /* maximum legal digit */

  switch(numberbase)
  { case 10:
     if(s[0]=='0')
     { if(tolower(s[1])=='x')
       { a=2;
        goto use_base_16;
       }
       else maxdigit='7'; /* for base 8 */
     }
     for(;s[a]>='0'&&s[a]<=maxdigit;a++);
    break;
    case 16:
    use_base_16:
     for(;isxdigit(s[a]);a++);
    break;
    default:
     fatalerror("internal inconsistency in tracking default number base");
  }
 return a;
}

long getnum(s,l) char *s;
{ long out;
  int x;
  char control[16]; /* must be [7+int(log10(l))]=10 (for string in sprintf) */

  switch(numberbase)
  { case 10:
      /* Does what you would want sscanf(s,"%*li",l,&out); if '*' did what
       * it does in printf. */
     sprintf(control,"%%%dli%%n",l); /* to "%XXli" where XX is value(l) */
     sscanf(s,control,&out,&x);
     if(x!=l) /* if sscanf processed a different # of chars then l chars. */
       warning("illegal character in number constant (or some other problem)");
    break;
    case 16:
     for(out=0;l--;s++) /* get value of hexidecimal number */
     { x= *s-'0';
       if(x>9) x+='0'-'a'+10;
       if(x<0||x>=16) fatalerror("illegal character in number const");
       out=(out<<4)+x;
     }
    break;
    default:
     fatalerror("internal inconsistency in tracking default number base");
  }
 return out;
}

powof2(x) unsigned long x;
{ int pow=0;
  while (1)
  { if(x&1&&x>1) return(-1);
    if(!x) return pow-1;
    x>>=1; pow++;
  }
}

fatalerror(s)
{ fprintf(stderr,"fatal error: %s: file %s line %d\n",s,infile,inline);
  fatal=1;
}

warning(s)
{ fprintf(stderr,"warning: %s: file %s line %d\n",s,infile,inline);
}

/* INSTRUCTION handling functions follow: */
inout(a) struct instructions *a; /* opcode for imm8,al form*/
/* handle input and output instructions */
{ struct operand op1,op2,top;
  op1=getoperand(',');
  if(!op1.lastop) op2=getoperand(',');
  else fatalerror("not enough operands");
  if(!op2.lastop) fatalerror("too many operands in/out");
  if(*a->arg==0xe4)
  { top=op1; op1=op2; op2=top; }
  if(op1.tp==TCONST)
  { if(op1.cval<0||op1.cval>0xff)
      fatalerror("in/out operand no good");
    else
    { if(op2.tp!=TREGMEM||op2.reg!=RAX)
        fatalerror("in/out operand 2 no good");
      prefix(0,0,op2.opsz!=S8&&op2.opsz!=defsz,0);
      switch(op2.opsz)
      { case S8:
         AB(*a->arg); AB(op1.cval); /* in/out al,imm8 */
        break;
        case S16o32:
         AB(*a->arg+1); AB(op1.cval); /* in/out (e)ax,imm8 */
        break;
        default: fatalerror("unknown size in inout");
      }
    }
  }
  else
  { if(op1.tp!=TREGMEM||op1.reg!=RDX||op1.opsz!=S16)
     fatalerror("in/out operand 1 no good. cutting inst. out");
    else
    { *a->arg+=8;
      if(op2.tp!=TREGMEM||op2.reg!=RAX)
       fatalerror("in/out operand 2 no good");
      prefix(0,0,op2.opsz!=S8&&op2.opsz!=defsz,0);
      switch(op2.opsz)
      { case S8:
         AB(*a->arg); /* in/out al,dx */
        break;
        case S16:
        case S32:
         AB(*a->arg+1); /* in/out (e)ax,dx */
        break;
        default: fatalerror("unknown size in inout");
      }
    }
  }
}

interrupt(a) struct instructions *a; /*none*/
/* handle interrupt instruction */
{ struct operand op;
  op=getoperand(',');
  if(!op.lastop||op.tp!=TCONST||op.cval<0||op.cval>0xff)
   fatalerror("interrupt operand no good");
  prefix(0,0,0,0);
  if(op.cval==3) AB(0xcc); /* int 3 */
  else if(op.cval==4) AB(0xce);
  else
  { AB(0xcd); AB(op.cval); } /* int ## */
}

imul(a) struct instructions *a; /* none*/
/* handle imul instruction */
/*MERE: not sure I like handle of 2 argument forms */
{ struct operand op1,op2,op3;
  int op1sz;
  op1=getoperand(',');
  if(op1.lastop)
  {singleop:
    if(op1.tp!=TREGMEM&&op1.tp!=TMEM)
      fatalerror("unknwon tp of opererator 2 imul");
    prefix(0,0,op1.opsz!=defsz,op1.tp==TMEM&&op1.addsz!=defsz);
    switch(op1.opsz)
    { case S8:
       AB(0xf6);
      break;
      case S16:
      case S32:
       AB(0xf7);
      break;
      default: fatalerror("unknown size of operand imul");
    }
    *op1.m.stf+=0x28;
    addfill(op1.m);
  }
  else
  { op2=getoperand(',');
    if(op1.tp!=TREGMEM||(op1.opsz!=S16&&op1.opsz!=S32))
      fatalerror("unknown type of operater 1 imul");
    if(op1.reg==RAX&&op2.tp!=TCONST)
    { op1sz=op1.opsz;
      op1=op2;
      if(op1.opsz==SUNKNOWN) op1.opsz=op1sz;
      else if(op1sz!=SUNKNOWN&&op1.opsz!=op1sz)
        fatalerror("1size mismatch");
      if(op1sz==SUNKNOWN||op1sz==S16o32) fatalerror("1.2size mismatch");
     goto singleop;
    }
    if(op2.lastop)
    { if(op2.tp==TCONST)
      { op3=op2;
        op2=op1;
       goto ops3;
      }
      else
      { prefix(0,0,op1.opsz!=defsz,op2.tp==TMEM&&op2.addsz!=defsz);
        if((op2.opsz!=SUNKNOWN&&op1.opsz!=op2.opsz)
           ||(op1.opsz!=S16&&op1.opsz!=S32))
          fatalerror("mismatch operand size");
        if(op2.tp!=TREGMEM&&op2.tp!=TMEM)
          fatalerror("illegal operand type 6");
        AB(0x0f); AB(0xaf);
        *op2.m.stf+=(op1.reg<<3);
        addfill(op2.m);
      }
    }
    else
    { op3=getoperand(',');
     ops3:
      if(!op3.lastop) fatalerror("too many operands imul");
      if(op3.tp!=TCONST||(op2.tp!=TREGMEM&&op2.tp!=TMEM))
        fatalerror("illegal operand type imul");
      if((op2.opsz!=SUNKNOWN&&op1.opsz!=op2.opsz)
         ||(op1.opsz!=S16&&op1.opsz!=S32))
        fatalerror("mismatch operand size");
      prefix(0,0,op1.opsz!=defsz,op2.tp==TMEM&&op2.addsz!=defsz);
      switch(op1.opsz)
      { case S8:
         AB(0x6b);
         *op2.m.stf+=(op1.reg<<3);
         addfill(op2.m);
         AB(op3.cval);
        break;
        case S16:
         AB(0x69);
         *op2.m.stf+=(op1.reg<<3);
         addfill(op2.m);
         AW(op3.cval);
        break;
        case S32:
         AB(0x69);
         *op2.m.stf+=(op1.reg<<3);
         addfill(op2.m);
         ADW(op3.cval);
        break;
        default: fatalerror("unknown operand size imul");
      }
    }
  }
}

enter(a) struct instructions *a; /* none*/
/* handle enter instruction */
{ struct operand op1,op2;
  op1=getoperand(',');
  if(op1.lastop||op1.tp!=TCONST||op1.cval<0||op1.cval>0xffff)
   fatalerror("enter operand 1 no good: cutting instr. out");
  else
  { op2=getoperand(',');
    if(!op2.lastop||op2.tp!=TCONST||op1.cval<0||op1.cval>0xff)
     fatalerror("enter operand 2 no good");
    prefix(0,0,0,0);
    AB(0xc8); AW(op1.cval); AB(op2.cval);
  }
}

jcc(a) struct instructions *a; /* code for imm8 form, no imm16 possible? */
/* handle jump conditionally instructions */
{ struct operand op1;
  int acopsz;
  op1=getoperand(',');
  prefix(0,0,pfx.opsz,0);
  acopsz=pfx.opsz^(defsz==S32)?S32:S16;
  if(op1.tp==TCONST&&(op1.sign=='+'||op1.sign=='-'))
  { makeshort();
    if((op1.cval<128&&op1.cval>=-128)&&op1.opsz!=SNEAR)
    { putshort(*a->arg,blkcnt,(outoff)op1.cval);
      if(op1.opsz!=SUNKNOWN)
        fatalerror("illegal jump size");
    }
    else
    { makenear(JTJcc,acopsz,blkcnt);
      putnear(JTJcc,*a->arg,acopsz,blkcnt,(outoff)op1.cval);
      if(op1.opsz!=SUNKNOWN&&op1.opsz!=SNEAR)
        fatalerror("illegal jump size");
    }
  }
  else if(*a->arg==0xe3||op1.jsz==SNEAR) /* jcxz */
  {/* must be particular size */
    makeshort();
    if((op1.tp==TMEM&&!op1.tplab)||(op1.tp!=TCONST&&op1.tp!=TMEM))
      fatalerror("unknown tp");
    putshort(*a->arg,blkcnt,(outoff)0);
    RAL(vxref,vxcnt+1,struct varxref);
    if(op1.jsz!=SNEAR)
    { if(op1.jsz!=SUNKNOWN)
        fatalerror("unknown jump size");
      op1.jsz=SSHORT;
    }
    else
    { if(*a->arg==0xe3) /*jcxz*/
        fatalerror("can't jcxz near");
      makenear(JTJcc,acopsz,blkcnt);
      putnear(JTJcc,*a->arg,acopsz,blkcnt,(outoff)0);
    }
    vxref[vxcnt].blk=blkcnt;
    vxref[vxcnt].off=blk[blkcnt].blen-1;
    vxref[vxcnt].line=inline;
/*HEREfuture: copy this to loop as well*/
    if(op1.m.jl.jloctp==JLFUNC)
    { vxref[vxcnt].tp=VTEXTERN;
      vxref[vxcnt].u.ext.ptrfnc=0;
      vxref[vxcnt].u.ext.addsz=acopsz;
      vxref[vxcnt].vnum=svars[op1.m.scoff].num;
      vxref[vxcnt].off=blk[blkcnt].blen-4;/*HEREfuture: might be 2?*/
    }
    else
    { vxref[vxcnt].tp=VTLABEL;
      vxref[vxcnt].u.lab.jloc=op1.m.jl;
      vxref[vxcnt].u.lab.opsz=acopsz;
      vxref[vxcnt].u.lab.isoff=1;
      vxref[vxcnt].u.lab.jsz=op1.jsz;
    }
    vxcnt++;
  }
  else
  { if((op1.tp==TMEM&&!op1.tplab)||(op1.tp!=TCONST&&op1.tp!=TMEM))
      fatalerror("unknown tp");
    blk[blkcnt].jloc=op1.m.jl;
    blk[blkcnt].jsz=SSHORT;
    blk[blkcnt].jtp=JTJcc;
    blk[blkcnt].jop= *a->arg;
    blk[blkcnt].opsz=acopsz;
    makeshort();
    blk[blkcnt].line=inline;
    blkcnt++;
    RAL(blk,blkcnt+1,struct outblock);
    AL(blk[blkcnt].bops,1,char);
    blk[blkcnt].blen=0;
    blk[blkcnt].inoff=blk[blkcnt-1].inoff+blk[blkcnt-1].blen;
  }
}

dblshift(a) struct instructions *a; /* code for immediate operand form */
/* handle double shift instructions*/
{ struct operand op1,op2,op3;
  op1=getoperand(',');
  if(op1.lastop) fatalerror("not enough operands");
  else
  { op2=getoperand(',');
    if(op2.lastop) fatalerror("not enough operands");
    else
    { op3=getoperand(',');
      if((op1.tp!=TREGMEM&&op1.tp!=TMEM)||op2.tp!=TREGMEM)
        fatalerror("unknown type of operand");
      if(op1.opsz!=SUNKNOWN&&op1.opsz!=op2.opsz)
        fatalerror("unknown operand size dblshift");
      if(op2.opsz!=S16&&op2.opsz!=S32)
        fatalerror("illegal operand size dblshift");
      prefix(0,0,op2.opsz!=defsz,op1.tp==TMEM&&op1.addsz!=defsz);
      AB(0x0f);
      switch(op3.tp)
      { case TCONST:
         AB(*a->arg);
         *op1.m.stf+=(op2.reg<<3);
         addfill(op1.m);
         AB(op3.cval);
        break;
        case TREGMEM:
         if(op3.reg!=RCX) fatalerror("illegal register dblshift");
         AB(*a->arg+1);
         *op1.m.stf+=(op2.reg<<3);
         addfill(op1.m);
        break;
        default: fatalerror("unknown type of operand shift double");
      }
    }
  }
}

shftrot(a) struct instructions *a; /* reg field */
/* handle shift and rotate instruction*/
{ struct operand op1,op2;
  op1=getoperand(',');
  if(op1.lastop) fatalerror("not enough operands shift");
  else
  { op2=getoperand(',');
    if(!op2.lastop) fatalerror("too many operands shift");
    if(op1.tp!=TMEM&&op1.tp!=TREGMEM) fatalerror("illegal type of operand");
    prefix(0,0,op1.opsz!=8&&op1.opsz!=defsz,op1.tp==TMEM&&op1.addsz!=defsz);
    *op1.m.stf+= *a->arg;
    switch(op2.tp)
    { case TCONST:
       if(op2.cval==1)
       { switch(op1.opsz)
         { case S8:
            AB(0xd0);
           break;
           case S16:
           case S32:
            AB(0xd1);
           break;
           default: fatalerror("1unknown operand size");
         }
         addfill(op1.m);
       }
       else
       { switch(op1.opsz)
         { case S8:
            AB(0xc0);
           break;
           case S16:
           case S32:
            AB(0xc1);
           break;
           default: fatalerror("2unknown operand size");
         }
         addfill(op1.m);
         AB(op2.cval);
       }
      break;
      case TREGMEM:
       if(op2.reg!=RCX) fatalerror("illegal register shift");
       switch(op1.opsz)
       { case S8:
          AB(0xd2);
         break;
         case S16:
         case S32:
          AB(0xd3);
         break;
         default: fatalerror("unknown operand size shift");
       }
       addfill(op1.m);
      break;
      default: fatalerror("unknown type of operand shift");
    }
  }
}

test(a) struct instructions *a; /*none*/
/* handle test instruction*/
{ struct operand op1,op2;
  op1=getoperand(',');
  if(op1.lastop) fatalerror("not enough operands");
  else
  { op2=getoperand(',');
    if(op1.tp==TREGMEM&&op1.reg==RAX&&op2.tp==TCONST)
    { prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,0);
      switch(op1.opsz)
      { case S8:
         AB(0xa8); AB(op2.cval);
        break;
        case S16:
         AB(0xa9); AW(op2.cval);
        break;
        case S32:
         AB(0xa9); ADW(op2.cval);
        break;
        default: fatalerror("3unknown operand size");
      }
    }
    else
    { if(op1.tp!=TREGMEM&&op1.tp!=TMEM)
        fatalerror("4unknown operand size");
      if(op1.opsz!=SUNKNOWN&&op2.tp!=TCONST&&op1.opsz!=op2.opsz)
        fatalerror("5unknown operand size");
      switch(op2.tp)
      { case TCONST:
         prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,op1.tp==TMEM
           &&op1.addsz!=defsz);
         switch(op1.opsz) /* all *op1.m.stf+=0 */
         { case S8:
            AB(0xf6);
            addfill(op1.m);
            AB(op2.cval);
           break;
           case S16:
            AB(0xf7);
            addfill(op1.m);
            AW(op2.cval);
           break;
           case S32:
            AB(0xf7);
            addfill(op1.m);
            ADW(op2.cval);
           break;
           default: fatalerror("6unknown operand size");
         }
        break;
        case TREGMEM:
         prefix(0,0,op2.opsz!=S8&&op2.opsz!=defsz,op1.tp==TMEM
           &&op1.addsz!=defsz);
         switch(op1.opsz)
         { case S8:
            AB(0x84);
           break;
           case S16:
           case S32:
            AB(0x85);
           break;
           default: fprintf(stderr,"unknwon operand size");
         }
         *op1.m.stf+=(op2.reg<<3);
         addfill(op1.m);
        break;
        default: fatalerror("7unknown operand type");
      }
    }
  }
}

xchg(a) struct instructions *a; /*none*/
/* handle xchg instruction*/
{ struct operand op1,op2,top;
  op1=getoperand(',');
  if(op1.lastop)
    fatalerror("need 2 operands to exchange them!");
  else
  { op2=getoperand(',');
    if(op1.tp==TMEM)
    { top=op1; op1=op2; op2=top; }
    if(op1.tp!=TREGMEM) fatalerror("one operand must be register");
    if(op1.opsz!=op2.opsz&&op2.opsz!=SUNKNOWN) fatalerror("missmatch size");
    switch(op2.tp)
    { case TREGMEM:
       if(op1.opsz==S16||op1.opsz==S32)
       { if(op2.reg==RAX)
         { AB(0x90+op1.reg);
          break;
         }
         else if(op1.reg==RAX)
         { AB(0x90+op2.reg);
          break;
         }
       }
      case TMEM:
       *op2.m.stf+=(op1.reg<<3);
       switch(op1.opsz)
       { case S8:
          AB(0x86);
          addfill(op2.m);
         break;
         case S16:
         case S32:
          AB(0x87);
          addfill(op2.m);
         break;
         default: fatalerror("unknown size xchg");
       }
      break;
      default: fatalerror("illegal operand type xchg");
    }
  }
}

memory2(a) struct instructions *a; /*2 bytes,reg field,permission*/
/* handle operations with single memory operand */
{ struct operand op1;
  op1=getoperand(',');
  if(!op1.lastop) fatalerror("too many operands");
  else
  { if(op1.opsz==SUNKNOWN)
    { if((a->arg[3]&OBT)==OBT)
        fatalerror("size can't be tracked");
      if(a->arg[3]&O16) op1.opsz=S16;
      else if(a->arg[3]&O32) op1.opsz=S32;
      else if(a->arg[3]&O8) op1.opsz=S8;
      else fatalerror("size untrackable");
    }
    else
    { if(a->arg[3]!=ONONE&&((a->arg[3]&OBT)!=OBT&&((a->arg[3]&O16
        &&op1.opsz!=S16)||(a->arg[3]&O32&&op1.opsz!=S32))))
       fatalerror("illegal operand size");
    }
    prefix(0,0,a->arg[3]!=ONONE&&op1.opsz!=defsz
      ,op1.tp==TMEM&&op1.addsz!=defsz);
    if(op1.tp!=TMEM&&op1.tp!=TREGMEM) fatalerror("1illegal operand type");
    AB(*a->arg); AB(a->arg[1]);
    *op1.m.stf+=a->arg[2];
    addfill(op1.m);
  }
}

muldiv(a) struct instructions *a; /*reg field*/
/* handle div and unsigned mul instruct */
{ struct operand op1;
  int op1sz;
  op1=getoperand(',');
  if(op1.tp==TREGMEM&&!(op1.lastop)&&op1.reg==RAX)
  { op1sz=op1.opsz;
    op1=getoperand(',');
  }
  else op1sz=0;
  if(!op1.lastop) fatalerror("too many operands");
  if(op1.opsz==SUNKNOWN) op1.opsz=op1sz;
  else if(op1sz&&op1.opsz!=op1sz)
    fatalerror("2size mismatch");
  prefix(0,0,op1.opsz!=defsz,op1.tp==TMEM&&op1.addsz!=defsz);
  if(op1.tp!=TMEM&&op1.tp!=TREGMEM) fatalerror("illegal type of reg");
  switch(op1.opsz)
  { case S8:
     AB(0xf6);
    break;
    case S16:
    case S32:
     AB(0xf7);
    break;
    default: fatalerror("unknwon size in opsz");
  }
  *op1.m.stf+= *a->arg;
  addfill(op1.m);
}

incdec(a) struct instructions *a; /*reg field*/
/* handle inc/dec instructions */
{ struct operand op1;
  op1=getoperand(',');
  if(!op1.lastop) fatalerror("too many operands");
  prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,op1.tp==TMEM&&op1.addsz!=defsz);
  switch(op1.opsz)
  { case S8:
     if(op1.tp!=TMEM&&op1.tp!=TREGMEM) fatalerror("illegal type of operand");
     AB(0xfe);
     *op1.m.stf+= *a->arg;
     addfill(op1.m);
    break;
    case S16:
    case S32:
     if(op1.tp==TMEM)
     { AB(0xff);
       *op1.m.stf+= *a->arg;
       addfill(op1.m);
     }
     else
     { if(op1.tp!=TREGMEM) fatalerror("illegal type of operand");
       AB(0x40+ *a->arg+ op1.reg);
     }
    break;
    default: fatalerror("8unknown operand size");
  }
}

byte(a) struct instructions *a; /* instruction's byte */
/* 1 byte instructions with fixed operands */
{ prefix(a->arg[1],0,0,0);
  AB(*a->arg);
}

dblbyte(a) struct instructions *a; /* both bytes */
/* 2 byte instructions with fixed operands */
{ prefix(0,0,0,0);
  AB(*a->arg); AB(a->arg[1]);
}

copbyte(a) struct instructions *a; /* byte,16/32bit */
/* 1 byte instructions with 16 or 32 bit operands */
{ prefix(a->arg[2],0,(defsz==S16&&!(a->arg[1]&O16))
    ||(defsz==S32&&!(a->arg[1]&O32)),pfx.addsz);
  AB(*a->arg);
}

general(a) struct instructions *a; /* reg field or base */
/*handles all the integer operations like add,sub,and, etc.*/
{ struct operand op1,op2;
  op1=getoperand(',');
  if(op1.lastop)
    fatalerror("not enough operands on general instruction");
  else
  { op2=getoperand(',');
    if(!op2.lastop) fatalerror("too many operands");
    switch(op1.tp)
    { case TMEM:
      memory:
       switch(op2.tp)
       { case TREGMEM:
          if((op1.opsz!=SUNKNOWN&&op1.opsz!=op2.opsz)||op2.opsz==SUNKNOWN)
            fatalerror("3size mismatch");
          prefix(0,0,op2.opsz!=S8&&op2.opsz!=defsz,op1.tp==TMEM
            &&op1.addsz!=defsz);
          *op1.m.stf+=(op2.reg<<3);
          if(op2.opsz==S8)
          { AB(*a->arg);
            addfill(op1.m);
          }
          else
          { if(op2.opsz!=S16&&op2.opsz!=S32) fatalerror("bad size");
            AB(*a->arg+1);
            addfill(op1.m);
          }
         break;
         case TCONST:
          if(op1.opsz==SUNKNOWN&&op2.opsz!=S16o32)
            fatalerror("unknown op size");
          prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,op1.tp==TMEM
            &&op1.addsz!=defsz);
          *op1.m.stf+= *a->arg;
         memCconst:
          if(op2.opsz==S8||op1.opsz==S8)
          { if(op1.opsz==S8) AB(0x80);
            else AB(0x83);
            addfill(op1.m);
            AB(op2.cval);
          }
          else
          { if(op2.opsz!=S16o32) fatalerror("unknown op 2 size");
            AB(0x81);
            addfill(op1.m);
            if(op1.opsz==S16) AW(op2.cval);
            else if(op1.opsz==S32) ADW(op2.cval);
            else fatalerror("op1.opsz bad");
          }
         break;
         default: fatalerror("unknown type of operand 2");
       }
      break;
      case TREGMEM:
       switch(op2.tp)
       { case TMEM:
          if((op2.opsz!=SUNKNOWN&&op2.opsz!=op1.opsz)||op1.opsz==SUNKNOWN)
            fatalerror("4size mismatch");
          prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,op2.addsz!=defsz);
          *op2.m.stf+=(op1.reg<<3);
          if(op1.opsz==S8)
          { AB(*a->arg+2);
            addfill(op2.m);
          }
          else
          { if(op1.opsz!=S16&&op1.opsz!=S32) fatalerror("bad size");
            AB(*a->arg+3);
            addfill(op2.m);
          }
         break;
         case TCONST:
          prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,0);
          *op1.m.stf+= *a->arg; /* used only if goto memCconst */
          if(op1.reg==RAX)
          { switch(op1.opsz)
            { case S8:
               AB(*a->arg+0x04);
               AB(op2.cval);
              break;
              case S16:
               AB(*a->arg+0x05);
               AW(op2.cval);
              break;
              case S32:
               if(op2.opsz==S8) goto memCconst; /*(uses less space)*/
               AB(*a->arg+0x05);
               ADW(op2.cval);
              break;
              default: fatalerror("unknwon operand size const");
            }
          }
          else goto memCconst;
         break;
         default: goto memory;
       }
      break;
      default: fatalerror("unknown type of operand 1");
    }
  }
}

bittest(a) struct instructions *a; /* 2nd byte; reg field is &0x38 */
/*handles bit test instructions with operands of  mem,(reg or const)*/
{ struct operand op1,op2;
  op1=getoperand(',');
  if(op1.lastop) fatalerror("not enough operands");
  else
  { op2=getoperand(',');
    switch(op2.tp)
    { case TREGMEM:
       if(op2.opsz!=S16&&op2.opsz!=S32)
         fatalerror("illegal operand size");
       if(op1.opsz!=SUNKNOWN&&op1.opsz!=op2.opsz)
         fatalerror("illegal operand size");
       prefix(0,0,op2.opsz!=defsz,op1.tp==TMEM&&op1.addsz!=defsz);
       if(op1.tp!=TMEM&&op1.tp!=TREGMEM)
         fatalerror("illegal operand size");
       AB(0x0f); AB(*a->arg);
       *op1.m.stf+=(op2.reg<<3);
       addfill(op1.m);
      break;
      case TCONST:
       if(op1.opsz==SUNKNOWN)
         fatalerror("9unknown operand size");
       prefix(0,0,op1.opsz!=defsz,op1.tp==TMEM&&op1.addsz!=defsz);
       if(op1.tp!=TMEM&&op1.tp!=TREGMEM)
         fatalerror("illegal operand size");
       AB(0x0f); AB(0xba);
       *op1.m.stf+= *a->arg&0x38;
       addfill(op1.m);
       AB(op2.cval);
      break;
      default: fatalerror("unknown operand type");
    }
  }
}

regmem2(a) struct instructions *a; /* use 2?, both bytes, perm1, perm2*/
/*handles 1or2 byte instructions which always have operands  reg,mem */
{ struct operand op1,op2;
  op1=getoperand(',');
  if(op1.lastop) fatalerror("not enough operands");
  else
  { op2=getoperand(',');
    if(!op2.lastop) fatalerror("too many operands");
    if(op1.tp!=TREGMEM||(op2.tp!=TMEM&&op2.tp!=TREGMEM)
       ||(op2.tp==TREGMEM&&!(a->arg[4]&RG)))
      fatalerror("illegal operand(s)");
    if(op2.opsz!=SUNKNOWN)
    { if(a->arg[4]!=ONONE&&(!!(a->arg[4]&OBT))==(op2.opsz==S8))
        fatalerror("5size mismatch");
      if(op1.opsz!=op2.opsz) fatalerror("opsize mismatch");
    }
    if(a->arg[3]!=ONONE&&(!!(a->arg[3]&OBT))==(op1.opsz==S8))
      fatalerror("6size mismatch");
    prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,op2.tp==TMEM&&op2.addsz!=defsz);
    if(*a->arg) AB(a->arg[1]);
    AB(a->arg[2]);
    *op2.m.stf+=(op1.reg<<3);
    addfill(op2.m);
  }
}

memreg(a) struct instructions *a; /* byte, perm1, perm2 */
/*handles instructions which always have operands  mem,reg */
{ struct operand op1,op2;
  op1=getoperand(',');
  if(op1.lastop) fatalerror("not enough operands");
  else
  { op2=getoperand(',');
    if(!op2.lastop) fatalerror("too many operands");
    if(op2.tp!=TREGMEM||(op1.tp!=TMEM&&op1.tp!=TREGMEM)
       ||(op1.tp==TREGMEM&&!(a->arg[2]&RG)))
      fatalerror("illegal operand(s)");
    if(op1.opsz!=SUNKNOWN)
    { if((a->arg[2]&OBT)==(op1.opsz==S8))
        fatalerror("7size mismatch");
    }
    if((a->arg[1]&OBT)==(op2.opsz==S8))
      fatalerror("8size mismatch");
    prefix(0,0,op2.opsz!=defsz,op1.tp==TMEM&&op1.addsz!=defsz);
    AB(*a->arg);
    *op1.m.stf+=(op2.reg<<3);
    addfill(op1.m);
  }
}

transfer(a) struct instructions *a; /* TRCLL or TRJMP, shortjmp or nearcall */
/*handles transfer instructions like call, jmp*/
{ struct operand op1;
  int acopsz;
  op1=getoperand(',');
  if(!op1.lastop) fatalerror("too many operands");
  prefix(0,0,pfx.opsz,0);
  acopsz=pfx.opsz^(defsz==S32)?S32:S16;
  if(op1.jsz!=SFAR&&*a->arg==TRCLL) op1.opsz=SNEAR;
  if((op1.tp==TMEM&&op1.tplab)||(op1.tp==TCONST&&op1.sign!='+'&&op1.sign!='-'))
  { if(op1.jsz==SFAR)
    { if((*a->arg)==TRCLL) AB(0x9a);
      else AB(0xea);
      if(op1.m.jl.jloctp==JLCONST)
      { if(acopsz==S32) ADW(op1.m.jl.off);
        else AW(op1.m.jl.off);
        AW(op1.seg);
      }
      else /* it has a label */
      { RAL(vxref,vxcnt+1,struct varxref);
        if(op1.m.jl.jloctp!=JLLABEL) warning("far JLEXTERN not implemented");
        vxref[vxcnt].blk=blkcnt;
        vxref[vxcnt].off=blk[blkcnt].blen;
        vxref[vxcnt].tp=VTLABEL;
        vxref[vxcnt].u.lab.jloc=op1.m.jl;
        vxref[vxcnt].u.lab.opsz=acopsz;
        vxref[vxcnt].line=inline;
        vxref[vxcnt].u.lab.isoff=0;
        vxref[vxcnt++].u.lab.jsz=SNEAR;
        if(acopsz==S32) ADW((long)0);
        else AW(0);
        AW(op1.seg);
      }
    }
    else if(op1.jsz==SNEAR&&op1.m.jl.jloctp==JLFUNC)
    { makeshort();
      makenear((*a->arg==TRCLL?JTCALL:JTJMP),acopsz,blkcnt);
      putnear(*a->arg==TRJMP?JTJMP:JTCALL,a->arg[1],acopsz,blkcnt
        ,(outoff)op1.m.jl.off);
      RAL(vxref,vxcnt+1,struct varxref);
      vxref[vxcnt].tp=VTEXTERN;
      vxref[vxcnt].u.ext.ptrfnc=0;
      vxref[vxcnt].u.ext.addsz=acopsz;
      vxref[vxcnt].vnum=svars[op1.m.scoff].num;
      vxref[vxcnt].blk=blkcnt;
      vxref[vxcnt].off=blk[blkcnt].blen-4;/*HEREfuture: might be 2?*/
      vxref[vxcnt++].line=inline;
    }
    else
    { makeshort();
      blk[blkcnt].jloc=op1.m.jl;
      blk[blkcnt].jop=a->arg[1];
      blk[blkcnt].jsz=SSHORT;
      blk[blkcnt].opsz=acopsz;
      if(*a->arg==TRJMP)
      { blk[blkcnt].jtp=JTJMP;
        if(op1.jsz==SNEAR) goto jmpnear;
      }
      else
      { blk[blkcnt].jtp=JTCALL;
       jmpnear:
        blk[blkcnt].jsz=SNEAR;
        makenear(blk[blkcnt].jtp,acopsz,blkcnt);
      }
      blk[blkcnt].line=inline;
      blkcnt++;
      RAL(blk,blkcnt+1,struct outblock);
      AL(blk[blkcnt].bops,1,char);
      blk[blkcnt].blen=0;
      blk[blkcnt].inoff=blk[blkcnt-1].inoff+blk[blkcnt-1].blen;
    }
  }
  else if(op1.tp==TCONST) /* + and - are here */
  { makeshort();
    if((op1.cval<128&&op1.cval>=-128)&&op1.opsz!=SNEAR)
    { putshort(a->arg[1],blkcnt,(outoff)op1.cval);
      if(op1.opsz!=SUNKNOWN)
        fatalerror("illegal jump size");
    }
    else
    { makenear(*a->arg==TRJMP?JTJMP:JTCALL,acopsz,blkcnt);
      putnear(*a->arg==TRJMP?JTJMP:JTCALL,a->arg[1],acopsz,blkcnt
        ,(outoff)op1.cval);
      if(op1.opsz!=SUNKNOWN&&op1.opsz!=SNEAR)
        fatalerror("illegal jump size");
    }
  }
  else if(op1.tp==TMEM||op1.tp==TREGMEM)
  { AB(0xff);
    if(*a->arg==TRJMP) *op1.m.stf+=0x20;
    else
    { *op1.m.stf+=0x10;
      if(*a->arg!=TRCLL) fatalerror("unknwon transfer type");
    }
    if(op1.jsz==SFAR) *op1.m.stf+=0x8;
    addfill(op1.m);
  }
  else fatalerror("unknown op tp for jmp");
}

loop(a) struct instructions *a; /* 1st byte */
/*handles loopcc instructions*/
{ struct operand op1;
  int acopsz;
  op1=getoperand(',');
  acopsz=pfx.opsz^(defsz==S32)?S32:S16;
  prefix(0,0,pfx.opsz,0);
  if(!op1.lastop) fatalerror("not enough operands");
  else
  {/* must be short loop */
    if(op1.jsz!=SNEAR&&op1.tp==TCONST&&(op1.sign=='+'||op1.sign=='-'))
    { makeshort();
      if((op1.cval<128&&op1.cval>=-128)&&op1.jsz!=SNEAR)
      { putshort(*a->arg,blkcnt,(outoff)op1.cval);
        if(op1.opsz!=SUNKNOWN)
          fatalerror("illegal jump size");
      }
      else
        fatalerror("loop near label");
    }
    else
    { makeshort();
      if((op1.tp==TMEM&&!op1.tplab)||(op1.tp!=TCONST&&op1.tp!=TMEM))
        fatalerror("unknown tp");
      putshort(*a->arg,blkcnt,(outoff)0);
      RAL(vxref,vxcnt+1,struct varxref);
      if(op1.m.jl.jloctp!=JLLABEL) warning("loop extern func not implemented");
      vxref[vxcnt].blk=blkcnt;
      vxref[vxcnt].tp=VTLABEL;
      vxref[vxcnt].off=blk[blkcnt].blen-1;
      vxref[vxcnt].u.lab.jloc=op1.m.jl;
      vxref[vxcnt].u.lab.isoff=1;
      vxref[vxcnt].u.lab.opsz=acopsz;
      vxref[vxcnt].line=inline;
      if(op1.jsz!=SUNKNOWN)
        fatalerror("illegal jump size");
      vxref[vxcnt++].u.lab.jsz=SSHORT;
    }
  }
}

mov(a) struct instructions *a; /*none*/
/*handles mov instructions*/
{ struct operand op1,op2;
  op1=getoperand(',');
  if(op1.lastop) fatalerror("one operand on mov?");
  else
  { op2=getoperand(',');
    if(!op2.lastop) fatalerror("too many operands in mov");
    switch(op1.tp)
    { case TMEM:
      memstuff:
       switch(op2.tp)
       { case TREGMEM:
          if((op1.opsz!=SUNKNOWN&&op1.opsz!=op2.opsz)||op2.opsz==SUNKNOWN)
            fatalerror("9size mismatch");
          prefix(0,0,op2.opsz!=S8&&op2.opsz!=defsz,op1.tp==TMEM
            &&op1.addsz!=defsz);
          *op1.m.stf+=(op2.reg<<3);
          if(op2.opsz==S8)
          { /* HEREfuture: this must be single byte offset like below*/
            AB(0x88);
            addfill(op1.m);
          }
          else
          { if(op2.opsz!=S16&&op2.opsz!=S32)
              fatalerror("unknown size");
            if(op2.reg==RAX&&((op1.m.sz==3&&op1.opsz==S16)||(op1.m.sz==5
              &&op1.opsz==S32))&&((*op1.m.stf==0x06&&op1.addsz==S16)
              ||(*op1.m.stf==0x05&&op1.addsz==S32)))
            { *op1.m.stf=0xa3;
              addfill(op1.m);
            }
            else
            { AB(0x89);
              addfill(op1.m);
            }
          }
         break;
         case TSEG:
          prefix(0,0,0,op1.tp==TMEM&&op1.addsz!=defsz);
          AB(0x8c);
          *op1.m.stf+=(op2.reg<<3);
          addfill(op1.m);
         break;
         case TCONST:
          prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,op1.addsz!=defsz);
          switch(op1.opsz)
          { case S8:
             AB(0xc6);
             addfill(op1.m);
             AB(op2.cval);
            break;
            case S16:
             AB(0xc7);
             addfill(op1.m);
             AW(op2.cval);
            break;
            case S32:
             AB(0xc7);
             addfill(op1.m);
             ADW(op2.cval);
            break;
            default: fatalerror("unknown opsz mov");
          }
         break;
         default: fatalerror("unknown type of 2nd operand");
       }
      break;
      case TREGMEM:
       switch(op2.tp)
       { case TCONST:
          prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,0);
          switch(op1.opsz)
          { case S8:
             AB(0xb0+op2.reg);
             AB(op2.cval);
            break;
            case S16:
             AB(0xb8+op2.reg);
             AW(op2.cval);
            break;
            case S32:
             AB(0xb8+op2.reg);
             ADW(op2.cval);
            break;
            case SUNKNOWN:
            default: fatalerror("unknown size");
          }
         break;
         case TMEM:
          if((op2.opsz!=SUNKNOWN&&op2.opsz!=op1.opsz)||op1.opsz==SUNKNOWN)
            fatalerror("10size mismatch");
          prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,op2.addsz!=defsz);
          *op2.m.stf+=(op1.reg<<3);
          if(op1.opsz==S8)
          { /* HEREfuture: this must be single byte offset like below*/
            AB(0x8a);
            addfill(op2.m);
          }
          else
          { if(op1.opsz!=S16&&op1.opsz!=S32)
              fatalerror("unknown size");
            if(op1.reg==RAX&&((op2.m.sz==3&&op2.opsz==S16)||(op2.m.sz==5
              &&op2.opsz==S32))&&((*op2.m.stf==0x06&&op2.addsz==S16)
              ||(*op2.m.stf==0x05&&op2.addsz==S32)))
            { *op2.m.stf=0xa1;
              addfill(op2.m);
            }
            else
            { AB(0x8b);
              addfill(op2.m);
            }
          }
         break;
         case TSPCR:
          prefix(0,0,0,0);
          AB(0x0f); AB(0x20);
          *op1.m.stf+=(op2.reg<<3);
          addfill(op1.m);
         break;
         case TSPDR:
          prefix(0,0,0,0);
          AB(0x0f); AB(0x21);
          *op1.m.stf+=(op2.reg<<3);
          addfill(op1.m);
         break;
         case TSPTR:
          prefix(0,0,0,0);
          AB(0x0f); AB(0x24);
          *op1.m.stf+=(op2.reg<<3);
          addfill(op1.m);
         break;
         default: goto memstuff;
       }
      break;
      case TSEG:
       if(op2.tp!=TMEM&&op2.tp!=TREGMEM) fatalerror("illegal operand 2 mov");
       prefix(0,0,0,op2.tp==TMEM&&op2.addsz!=defsz);
       AB(0x8e);
       *op2.m.stf+=(op1.reg<<3);
       addfill(op2.m);
      break;
      case TSPCR:
       if(op2.tp!=TREGMEM) fatalerror("illegal operand 2 mov");
       prefix(0,0,0,0);
       AB(0x0f); AB(0x22);
       *op2.m.stf+=(op1.reg<<3);
       addfill(op2.m);
      break;
      case TSPDR:
       if(op2.tp!=TREGMEM) fatalerror("illegal operand 2 mov");
       prefix(0,0,0,0);
       AB(0x0f); AB(0x23);
       *op2.m.stf+=(op1.reg<<3);
       addfill(op2.m);
      break;
      case TSPTR:
       if(op2.tp!=TREGMEM) fatalerror("illegal operand 2 mov");
       prefix(0,0,0,0);
       AB(0x0f); AB(0x26);
       *op2.m.stf+=(op1.reg<<3);
       addfill(op2.m);
      break;
      default: fatalerror("illegal operand type in mov");
    }
  }
}

movxt(a) struct instructions *a; /*none*/
/*handles mov with sign or zero extension instructions*/
{ struct operand op1,op2;
  op1=getoperand(',');
  if(op1.opsz!=S32&&op1.opsz!=S16)
    fatalerror("illegal operand size movxt");
  if(op1.lastop) fatalerror("not enough operands movxt");
  else
  { op2=getoperand(',');
    if(op1.tp!=TREGMEM||(op2.tp!=TREGMEM&&op2.tp!=TMEM))
      fatalerror("illegal operand type movxt");
    if(!op2.lastop) fatalerror("too many operands movxt");
    prefix(0,0,op1.opsz!=defsz,op2.tp==TMEM&&op2.addsz!=defsz);
    AB(0x0f);
    switch(op2.opsz)
    { case S8:
       AB(*a->arg);
      break;
      case S16:
       if(op1.opsz!=S32) fatalerror("illegal opsz combo movxt");
       AB(*a->arg+1);
      break;
      default: fatalerror("unknown operand size movxt");
    }
    *op2.m.stf+=(op1.reg<<3);
    addfill(op2.m);
  }
}

negnot(a) struct instructions *a; /* reg field */
/* handle neg and not instructions */
{ struct operand op1;
  op1=getoperand(',');
  if(!op1.lastop) fatalerror("too many operands neg/not");
  if(op1.tp!=TREGMEM&&op1.tp!=TMEM) fatalerror("unknown type of operand");
  prefix(0,0,op1.opsz!=S8&&op1.opsz!=defsz,op1.tp==TMEM&&op1.addsz!=defsz);
  switch(op1.opsz)
  { case S8:
     AB(0xf6);
    break;
    case S16:
    case S32:
     AB(0xf7);
    break;
    default: fatalerror("unknown operand size neg/not");
  }
  *op1.m.stf+= *a->arg;
  addfill(op1.m);
}

pushpop(a) struct instructions *a; /*offset(0/8)for push or pop regs*/
/*handles push and pop (not pusha;pushf) instructions */
{ struct operand op1;
  int argval;
  op1=getoperand(',');
  if(!op1.lastop) fatalerror("too many operands");
  else
  { switch(op1.tp)
    { case TCONST:
       if(*a->arg) fatalerror("can't pop to a constant");
       prefix(0,0,pfx.opsz,0);
       switch(op1.opsz)
       { case S8:
          AB(0x6a);
          AB(op1.cval);
         break;
         case S16o32:
          AB(0x68);
          if((defsz==S16&&!pfx.opsz)||(defsz==S32&&pfx.opsz)) AW(op1.cval);
          else ADW(op1.cval);
         break;
         default: fatalerror("unknown size of operand push");
       }
      break;
      case TREGMEM:
       if(op1.opsz!=S16&&op1.opsz!=S32) fatalerror("unknown size push/pop");
       prefix(0,0,op1.opsz!=defsz,0);
       AB(0x50+*a->arg+op1.reg);
      break;
      case TMEM:
       if(op1.opsz==S8) fatalerror("can't push/pop memory byte");
       prefix(0,0,(op1.opsz!=SUNKNOWN&&op1.opsz!=defsz)
          ||(op1.opsz==SUNKNOWN&&pfx.opsz),op1.addsz!=defsz);
       if(*a->arg) AB(0x8f);
       else
       { AB(0xff);
         *op1.m.stf+=0x30;
       }
       addfill(op1.m);
      break;
      case TSEG:
       argval= *a->arg>>3;
       prefix(0,0,0,0);
       switch(op1.reg)
       { case RCS:
          if(argval) warning("pop cs ;code exists only on 8086");
          AB(0x0e+argval);
         break;
         case RSS:
          AB(0x16+argval);
         break;
         case RDS:
          AB(0x1e+argval);
         break;
         case RES:
          AB(0x06+argval);
         break;
         case RFS:
          AB(0x0f); AB(0xa0+argval);
         break;
         case RGS:
          AB(0x0f); AB(0xa8+argval);
         break;
         default: fatalerror("unknown segment register");
       }
      break;
      default: fatalerror("unknown type of operand push/pop");
    }
  }
}

returnproc(a) struct instructions *a; /* near,imm16 form */
/* handle return instructions */
{ struct operand op1;
  int ch;
  do
  { ch=getic();
  } while(ch==' '||ch=='\t');
  ungetic(ch);
  prefix(0,0,pfx.opsz,0);
  if(isdigit(ch))
  { op1=getoperand(',');
    if(!op1.lastop) fatalerror("too many operands return");
    if(op1.tp!=TCONST) fatalerror("unknown tp of operand");
    AB(*a->arg); AW(op1.cval);
  }
  else
  { AB(*a->arg+1);
  }
}

typeget()
{ int b=0;
  char str[STRINGSZ];
  typegotsize=0;
 unsignedredo:
  getstr(str,' ',0); /* type */
  switch(strlen(str))
  { case 3:
     if(!strcmp(str,"int")) b+=SINT;
     else goto badtype;
    break;
    case 4:
     if(!strcmp(str,"char")) b+=SCHAR;
     else if(!strcmp(str,"long")) b+=SLONG;
     else if(!strcmp(str,"func"))
     { if(b) warning("no such thing as an unsigned func");
       b=SFUNC;
     }
     else goto badtype;
    break;
    case 5:
     if(!strcmp(str,"array"))
     { if(b) warning("no such thing as unsigned array");
       b=SARRAY;
       typegotsize=1;
       getstr(str,' ',0); /* type */
       typesize=getnum(str,strlen(str));
     }
     else goto badtype;
    break;
    case 6:
     if(!strcmp(str,"struct"))
     { if(b) warning("no such thing as an unsigned struct");
       b=SSTRUCT;
       typegotsize=1;
       getstr(str,' ',0); /* type */
       typesize=getnum(str,strlen(str));
     }
     else if(!strcmp(str,"double"))
     { if(b) warning("no such thing as an unsigned double");
       b=SDOUBLE;
     }
     else goto badtype;
    break;
    case 7:
     if(!strcmp(str,"pointer"))
     { if(b) warning("no such thing as an unsigned pointer");
       b=SPOINTER;
       typegotsize=1;
       getstr(str,' ',0); /* type */
       typesize=getnum(str,strlen(str));
     }
     else goto badtype;
    break;
    case 8:
     if(!strcmp(str,"unsigned"))
     { if(b) warning("multiple unsigneds");
       b=SUNSIGNED;
      goto unsignedredo;
     }
    break;
    badtype:
    default:
     fatalerror("unknown function type");
  }
 return b;
}

/* COPROCESSOR related INSTRUCTIONS */
memory(a) struct instructions *a; /* byte1, regfld, f2(opsz), f2(regAX) */
/* handle single op memory access instructions; only f instr. use currently */
{ struct operand op1;
  op1=getoperand(',');
  if(!op1.lastop) fatalerror("too many operands memory");
  else
  { if(op1.tp==TREGMEM&&!strcmp(a->nm,"fstsw"))
    { if(op1.reg!=RAX) fatalerror("illegal dest. reg for fstsw");
      prefix(0,0,0,0);
      AW(0xe0df);
    }
    else
    { prefix(0,0,pfx.opsz,op1.addsz!=defsz);
      if(pfx.opsz&&!a->arg[2]) warning("not needed operandsize override");
      if((op1.tp!=TREGMEM&&(!a->arg[3]||!op1.reg))&&op1.tp!=TMEM)
        fatalerror("2illegal operand type");
      AB(*a->arg);
      *op1.m.stf+=a->arg[1];
      addfill(op1.m);
    }
  }
}

escape(a) struct instructions *a; /* none */
/* handle esc opcode,mem   instruction*/
{ struct operand op1,op2;
  op1=getoperand(',');
  if(op1.lastop) fatalerror("not enough operands esc");
  else
  { op2=getoperand(',');
    if(!op1.lastop) fatalerror("too many operands esc");
    if(op1.tp!=TCONST||(op2.tp!=TREGMEM&&op2.tp!=TMEM))
      fatalerror("illegal operand type(s) esc");
    prefix(0,0,pfx.opsz,op2.addsz!=defsz);
    AB(0xd8|((op1.cval>>3)&7));
    *op2.m.stf+=((op1.cval&7)<<3);
    addfill(op2.m);
  }
}

fintern(a) struct instructions *a; /* op1,op2, flag(is "...,st" ok?) */
/* handle instructions that can only index 387's internal regs st(0-7) */
{ struct operand op1,op2;
  op1=getoperand(',');
  if(!op1.lastop)
  { op2=getoperand(',');
    if(!op2.lastop||!a->arg[2])
      fatalerror("too many operands finternal");
    if(op2.tp!=TMATH||op2.reg)
      fatalerror("illegal 2nd operand finternal");
  }
  if(op1.tp!=TMATH)
    fatalerror("illegal 1st operand finternal");
  prefix(0,0,0,0);
  AB(*a->arg); AB(a->arg[1]+op1.reg);
}

fgenops(a) struct instructions *a; /* reg field, flag(MINT or MFLT) */
/* handle add,mul,com,comp,subsubr,div,divr for 387 */
{ struct operand op1,op2;
  if(lastinstrend!='\n')
  { op1=getoperand(',');
    if(op1.lastop) /* one operand */
    { prefix(0,0,0,op1.tp==TMEM&&op1.addsz!=defsz);
      switch(op1.tp)
      { case TMEM:
         switch(op1.opsz)
         { case S16:
            if(a->arg[1]!=MINT) fatalerror("16 bit arg must be int/not float");
            AB(0xde);
            *op1.m.stf+= *a->arg; addfill(op1.m);
           break;
           case S32:
            switch(a->arg[1])
            { case MINT: AB(0xda); break;
              case MFLT: AB(0xd8); break;
              default: fatalerror("internal error fgenops");
            }
            *op1.m.stf+= *a->arg; addfill(op1.m);
           break;
           case S64:
            if(a->arg[1]!=MFLT) fatalerror("no long integer fops");
            AB(0xd8);
            *op1.m.stf+= *a->arg; addfill(op1.m);
           break;
           case S48: /*THESE are errors, just make clear they exist*/
           case S80:
           default: fatalerror("unknown size in fgenops");
         }
        break;
        case TMATH:
         if(*a->arg==0x10||*a->arg==0x18)
         { AB(0xd8); AB(0xc0+*a->arg+op1.reg); } /*fcom(p) st,st(reg)*/
        break;
        default: fatalerror("unknown type of operand");
      }
    }
    else /* two operands */
    { op2=getoperand();
      prefix(0,0,0,0);
      if(a->arg[1]!=MFLT||op1.tp!=TMATH||op2.tp!=TMATH)
        fatalerror("illegal operand(s) fgenops");
      else
      { if(op1.reg)
        { if(*a->arg==0x10||*a->arg==0x18)
            fatalerror("fcom(p) st(i),st");
          if(*a->arg>0x18) *a->arg^=8; /* if div(r) or sub(r) */
          if(op2.reg) fatalerror("one operand must be st(0)");
          AB(0xdc); AB(0xc0+*a->arg+op1.reg); /* foper st(i),st */
        }
        else
        { AB(0xd8); AB(0xc0+*a->arg+op2.reg); /* foper st,st(i) */ }
      }
    }
  }
  else /* no operands */
  { if(a->arg[1]!=MFLT) fatalerror("fi(oper) needs operands");
    prefix(0,0,0,0);
    if(*a->arg==0x10||*a->arg==0x18)
    { AB(0xd8); AB(0xc1+*a->arg); } /*fcom(p) st,st(1)*/
    else
    { AB(0xde); AB(0xc1); } /*foperp st(1),st*/
  }
}

fst_ld(a) struct instructions *a; /* (reg field)&18, flag(MINT or MFLT) */
/* handle fld/fild/fst/fstp/fist/fistp*/
{ struct operand op1;
  op1=getoperand();
  switch(op1.tp)
  { case TMEM:
     prefix(0,0,0,op1.addsz!=defsz);
     switch(op1.opsz)
     { case S16:
        if(a->arg[1]!=MINT) fatalerror("internal error fst_ld");
        AB(0xdf); /*fi[st/stp/ld]*/
        *op1.m.stf+= *a->arg; addfill(op1.m);
       break;
       case S32:
        switch(a->arg[1])
        { case MINT: AB(0xdb); break;
          case MFLT: AB(0xd9); break;
          default: fatalerror("internal error fst_ld2");
        }
        *op1.m.stf+= *a->arg; addfill(op1.m);
       break;
       case S64:
        switch(a->arg[1])
        { case MINT:
           AB(0xdf);
           *op1.m.stf+= fsthighpert(*a->arg);
          break;
          case MFLT:
           AB(0xdd); *op1.m.stf+= *a->arg;
          break;
          default: fatalerror("internal error fst_ld3");
        }
        addfill(op1.m);
       break;
       case S80:
        if(a->arg[1]!=MFLT) fatalerror("internal erro fst_ld4");
        AB(0xdb);
        *op1.m.stf+= fsthighpert(*a->arg); addfill(op1.m);
       break;
       default: fatalerror("unknown operand size");
     }
    break;
    case TMATH:
     prefix(0,0,0,0);
     if(a->arg[1]!=MFLT) fatalerror("internal error fst_ld");
     switch(*a->arg)
     { case 0x0: /* fld */
        AB(0xd9); AB(0xc0+op1.reg); /*fld st(i)*/
       break;
       case 0x10: /* fst */
       case 0x18: /* fstp */
        AB(0xdd); AB(0xc0+*a->arg+op1.reg);
       break;
       default: fatalerror("internal error fst_ld 10");
     }
    break;
    default: fatalerror("unknown type of operand fst_ld");
  }
}

fsthighpert(a)
/* translates in certain way fst/fld/fstp #'s higher mod byte */
{ switch(a)
  { case 0x0: return 0x28;
    case 0x18: return 0x38;
    default: fatalerror("internal error fsthighpert");
  }
 return 0;
}

/* PSEUDO INSTRUCTION HANDLING FUNCTIONS */

modes(a) struct instructions *a; /* none */
/*modes [expseg/noexpseg] [16bit/32bit] [expsz/noexpsz]
 *modes [86/286/386] [jccnear/nojccnear] /*HEREfuture these not fully
 *   impleted imm
 *modes [base10/base16]
 */
{ char temp[STRINGSZ];
  int ch,len;
  do
  { ch=getstr(temp,' ',0);
    len=strlen(temp);
    switch(len) /* convenient method of reducing exection time */
    { case 2: /* 86 */
       if(!strcmp(temp,"86")) prst=RS86;
       else break;
      continue;
      case 3: /* 286/386 */
       if(!strcmp(temp,"286")) prst=RS286;
       else if(!strcmp(temp,"386")) prst=RS386;
       else break;
      continue;
      case 5: /* 16bit/32bit/expsz */
       if(!strcmp(temp,"16bit")) defsz=S16;
       else if(!strcmp(temp,"32bit")) defsz=S32;
       else if(!strcmp(temp,"expsz")) expsz=1;
       else break;
      continue;
      case 6: /* expseg */
       if(!strcmp(temp,"expseg")) expseg=1;
       else if(!strcmp(temp,"base10")) numberbase=10;
       else if(!strcmp(temp,"base16")) numberbase=16;
       else break;
      continue;
      case 7: /* noexpsz/jccnear */
       if(!strcmp(temp,"noexpsz")) expsz=0;
       else if(!strcmp(temp,"jccnear")) jccnear=1;
       else break;
      continue;
      case 8: /* noexpseg */
       if(!strcmp(temp,"noexpseg")) expseg=0;
       else break;
      continue;
      case 9: /* nojccnear */
       if(!strcmp(temp,"nojccnear")) jccnear=0;
       else break;
      continue;
    } /* execution gets beyond here only on break or default */
    fatalerror("unrecognised modes operand");
  } while(ch!='\n'&&ch!=EOF);
}

datab(a) struct instructions *a; /* none */
{ struct operand op1;
  int b;
  if(lastinstrend!='\n')
   do
   { op1=getoperand(' ');
     switch(op1.tp)
     { case TSTRING:
        for(b=0;op1.string[b];b++) AB(op1.string[b]);
       break;
       case TCONST:
        AB(op1.cval);
       break;
       default: fatalerror("unknown tp of operand in db");
     }
   } while(!op1.lastop);
}

dataw(a) struct instructions *a; /* none */
{ struct operand op1;
  int b;
  if(lastinstrend!='\n')
   do
   { op1=getoperand(' ');
     switch(op1.tp)
     { case TSTRING:
        for(b=0;op1.string[b];b++) AB(op1.string[b]);
       break;
       case TMEM:
        if(op1.m.sz!=3||*op1.m.stf!=6||op1.m.addsz!=S16)
          fatalerror("illegal address stuff in dw");
        else
        { if(outplace)
          { /*HEREfuture: for 8086 do prexref */
          }
          else
          { memmove(op1.m.stf,op1.m.stf+1,op1.m.sz-1);
            op1.m.sz--; op1.m.xoff--;
if(op1.m.xoff) warning("xoff error");
            addfill(op1.m);
          }
        }
       break;
       case TCONST:
        AW(op1.cval);
       break;
       default: fatalerror("unknown tp of operand in dw");
     }
   } while(!op1.lastop);
}

datadw(a) struct instructions *a; /* none */
{ struct operand op1;
  int b;
  if(lastinstrend!='\n')
   do
   { op1=getoperand(' ');
     switch(op1.tp)
     { case TSTRING:
        for(b=0;op1.string[b];b++) AB(op1.string[b]);
       break;
       case TMEM:
        if(op1.m.sz!=5||*op1.m.stf!=5||op1.m.addsz!=S32)
          fatalerror("illegal address stuff in ddw");
        else
        { if(outplace)
          { struct prexref *px;
            switch(outplace)
            { default: fatalerror("unknown outplace in datadw");
               /* error recovery: make px semi-legitimate pointer */
              case PCEXTERN:
               RAL(extxref,extxcnt+1,struct prexref);
               px=extxref+extxcnt++;
               px->eoff=lastextnum;
               px->poff=exttbl[lastextnum].presz;
            }
            px->soff=svars[op1.m.scoff].u.off;
            /*above applies for both static & extern*/
            switch(op1.m.xexist)
            { case VTEXTERN:
               px->sect=SCPUBLIC;
              break;
              default: fatalerror("illegal preinitialization");
            }
          }
          else
          { memmove(op1.m.stf,op1.m.stf+1,op1.m.sz-1);
            op1.m.sz--; op1.m.xoff--;
if(op1.m.xoff) warning("xoff error");
            addfill(op1.m);
          }
        }
       break;
       case TCONST:
        ADW(op1.cval);
       break;
       default: fatalerror("unknown tp of operand in ddw");
     }
   } while(!op1.lastop);
}

absvardef(a) struct instructions *a; /*none*/
/* declare(not define) absolute variables */
{ char temp[STRINGSZ];
  int c,scan;
  do
  { c=getstr(temp,' ',0);
    scan=labeler(temp,strlen(temp));
    if(svars[scan].here)
    { if(svars[scan].tp!=VTABSOLUTE)
        fatalerror("declare non-absolute variable as absolute");
      else
        warning("already defined variable being declared");
    }
    else
    { if(svars[scan].num!=varcnt-1)
      { if(svars[scan].tp==VTABSOLUTE)
          warning("variable already declared");
        else
          fatalerror("absolute variable already declared as something else");
      }
      svars[scan].tp=VTABSOLUTE; /* all other variables already handled */
      svars[scan].here=0;
    }
  } while(c!='\n'&&c!=EOF);
}

funcinit(a) struct instructions *a;/* OFUNC or OSTATFUNC */
{ int b;
  char temp[STRINGSZ];
  if(mode!=MD386) fatalerror("can't init func in mode other then MD386/86");
  if(blkcnt||blk->blen) fatalerror("junk inserted before func definition");
/* FUNCtp(1),name(?),0(1),tp(1),size(0or4) */
  AB(*a->arg);
  getstr(temp,' ',0); /* name */
  b=0;
  do AB(temp[b]); while(temp[b++]);
  AB(typeget());
  if(typegotsize) ADW(typesize);
  b=labeler(temp,strlen(temp));
  svars[b].tp=VTEXTERN;
  svars[b].u.off=exttblsz;
  if(svars[b].here)
    fatalerror("already defined variable defined as function");
  svars[b].here=1;
  svars[b].local=0;
  RAL(exttbl,exttblsz+1,struct exttbl);
  Scp(&exttbl[exttblsz].nm,temp);
  exttbl[exttblsz].tp=SFUNC;
  exttbl[exttblsz].scope=(*a->arg)==OSTATFUNC;
  exttbl[exttblsz++].prepres=0;
}

endfunc(a) struct instructions *a;/*none*/
{ long b;
  int dst;
  xreffill(1);
 /* write total block size */
  b=blk[blkcnt].inoff+blk[blkcnt].blen;
  if(!b) fatalerror("function not started");
  fwrite(&b,sizeof(long),1,outfile);
 /* write blocks */
  for(b=0;b<=blkcnt;b++)
  { fwrite(blk[b].bops,1,blk[b].blen,outfile);
    free(blk[b].bops);
  }
 /* write 12 0's */
  fwrite(&exrcnt,4,1,outfile);
  fwrite(exref,sizeof(struct externxref),exrcnt,outfile);
 /*reset junk*/
  exrcnt=stkoff=stkptrexist=stkvcnt=blkcnt=vxcnt=0;
  stkvAL=BUFSZ;
  RAL(exref,1,struct externxref);
  RAL(blk,1,struct outblock);
  RAL(vxref,1,struct varxref);
  for(b=dst=0;b<varcnt;b++) if(!svars[b].local) svars[dst++]=svars[b];
  varcnt=dst;
  RAL(svars,varcnt,struct variables);
  RAL(stkvars,BUFSZ,struct stackvars);
  AL(blk->bops,1,char);
  blk->blen=0;
}

stackvars(a) struct instructions *a; /*none*/
/* define stack vars backwards (see below stackptr)*/
{ char temp[STRINGSZ];
  int ch,len;
  long num;
  if(lastinstrend!='\n') do
  { ch=getstr(temp,' ',0);
    len=strlen(temp)+1;
    if(stkvcnt>=stkvAL)
    { stkvAL=stkvcnt+BUFSZ;
      RAL(stkvars,stkvAL,struct stackvars);
    }
    AL(stkvars[stkvcnt].nm,len,char);
    strcpy(stkvars[stkvcnt].nm,temp,len);
    ch=getstr(temp,' ',0);
    len=strlen(temp);
    num=getnum(temp,len);
    stkoff-=num;
    stkvars[stkvcnt++].off=stkoff;
  } while(ch!='\n'&&ch!=EOF);
  if(stkptrexist) definestackvars();
}

stackptr(a) struct instructions *a; /*none*/
/* stackvars a 1
 * stackptr
 * stackvars b 2
 *  above makes assembler assume that bp points to a, and bp-2 points to b */
{ stkptrexist=1;
  stkptr=stkoff;
  definestackvars();
}

definestackvars()
{ int a;
  varcnttp vr;
  for(a=0;a<stkvcnt;a++)
  { vr=labeler(stkvars[a].nm,strlen(stkvars[a].nm));
    if(svars[vr].num!=varcnt-1||svars[vr].here)
      fatalerror("already defined variable defined as stack");
    free(stkvars[a].nm);
    svars[vr].tp=VTSTACK;
    svars[vr].here=1;
    svars[vr].u.off=stkvars[a].off-stkptr;
  }
  stkvAL=BUFSZ;
  stkvcnt=0;
  RAL(stkvars,stkvAL,struct stackvars);
}

extvardef(arg) struct instructions *arg; /*none*/
/* extern
 *  name tp sz [prepres]
 *  db/dw  appropriate args until end of line;[only if last was prepres]
 *  name tp sz 1 d? "                      "
 * endextern
 */
{ char temp[STRINGSZ];
  int ch,len;
  varcnttp lab;

  outplace=PCEXTERN;
  if(mode!=MD386) fatalerror("init. extern variables without mode MD386");
  lastextpres=0;
  do
  { lastinstrend=ch=getstr(temp,' ',1);
    len=strlen(temp);
    if(!len) fatalerror("unexpected no lenghth string");
    if(len==2)
    { if(!strcmp(temp,"db"))
      { if(lastextpres) datab(arg);
        else fatalerror("init length extern");
      }
      else if(!strcmp(temp,"dw"))
      { if(lastextpres) dataw(arg);
        else fatalerror("init length extern");
      }
      else goto isname;
      if(ch==EOF) fatalerror("unexpected end of input");
    }
    else if(len==3&&!strcmp(temp,"ddw"))
    { if(lastextpres) datadw(arg);
      else fatalerror("init length extern");
      if(ch==EOF) fatalerror("unexpected end of input");
    }
    else
    {isname:
      if(lastextpres&&exttbl[lastextnum].presz<exttbl[lastextnum].sz)
      { while(exttbl[lastextnum].presz<exttbl[lastextnum].sz)
          exttbl[lastextnum].prestf[exttbl[lastextnum].presz++]=0;
      }
      if(len==9&&!strcmp(temp,"endextern")) break;
      lab=labeler(temp,len);
      if(svars[lab].num!=varcnt-1||svars[lab].here)
        fatalerror("variable already defined redefined as extern");
      if(!blkcnt&&!blk->blen) svars[lab].local=0;
      RAL(exttbl,exttblsz+1,struct exttbl);
      Scp(&exttbl[exttblsz].nm,svars[lab].nm);
      svars[lab].tp=VTEXTERN;
      lastextnum=svars[lab].u.off=exttblsz;
      if(ch==EOF) fatalerror("unexpected end of input");
      exttbl[exttblsz].tp=typeget();
      exttbl[exttblsz].auxsz=typesize;
      if(exttbl[exttblsz].tp!=SFUNC) svars[lab].here=1;
      if(!typegotsize)
      { ch=getstr(temp,' ',0);
        len=strlen(temp);
        if(!len) fatalerror("unknown string");
        exttbl[exttblsz].sz=getnum(temp,len);
      }
      else if(exttbl[exttblsz].tp==SPOINTER) exttbl[exttblsz].sz=4;
      else exttbl[exttblsz].sz=typesize; /*probably SSTRUCT,SARRAY*/

      ch=getstr(temp,' ',0);
      len=strlen(temp);
      if(!len) fatalerror("unknown string");
      lastextpres=exttbl[exttblsz].prepres=getnum(temp,len);

      exttbl[exttblsz].presz=0;
      if(lastextpres)
        AL(exttbl[exttblsz].prestf,exttbl[exttblsz].sz,char);

      ch=getstr(temp,' ',0);
      if(!strcmp(temp,"public")) exttbl[exttblsz].scope=0; /* SCPEXTERN */
      else if(!strcmp(temp,"static")) exttbl[exttblsz].scope=1; /* SCPSTATIC */
      else fatalerror("unknown range");

      exttblsz++;
    }
  } while(ch!=EOF);
  if(lastextpres&&exttbl[lastextnum].presz<exttbl[lastextnum].sz)
  { while(exttbl[lastextnum].presz<exttbl[lastextnum].sz)
      exttbl[lastextnum].prestf[exttbl[lastextnum].presz++]=0;
  }
  outplace=PCBLOCKS;
}
