/* malloc.c - high performance malloc without bug in UNIX's high performance
 *  version of malloc that occasionally causes core dumps.
 * NOTE: modified to replace UNIX's malloc, originally from DOS C compiler
 */
#include <stdio.h>

char *sbrk();
int brk();

#define SBRKSZ 8192

/*#include <ctype.h>
#define EOF (-1)
*/

/*#define TESTING /* */
#ifdef TESTING
# define LBERROR(args,func) do  \
   { fprintf  args;             \
     func;                      \
   } while(0)
#else
# define LBERROR(str,func) func
/*# define OPTIMIZED
*/
#endif

/* memory allocation junk. */

static char spacefound;
#define MALLOC_BASIC 1
#define MALLOC_BTREE 2

static char *maxposition,*firstchunk;
static char *lastchunk;
/* each chunk allocated as follows:
 *size          4  ; SIZEPIECE
 *pointertonext 4  ; NEXTPIECE
 *pointertoprev 4  ; PREVPIECE
 *userspace     sz ; located at byte offset HEADSZ
 */
#define SIZEPIECE(ptr) (*((int *)(ptr)))
#define NEXTPIECE(ptr) (((char **)(ptr))[1])
#define PREVPIECE(ptr) (((char **)(ptr))[2])
#define HEADSZ 12
#define NULL 0

/*FILE *stderr;
printmalloc()
{ char *chunk;
  int a;
fprintf(stderr,"firstchunk %#x alloccnt %d\n",firstchunk,alloccnt);
  for(chunk=firstchunk,a=0;a<=alloccnt;a++,chunk=NEXTPIECE(chunk))
    fprintf(stderr,"a %d chunk %#x NEXTPIECE %#x PREVPIECE %#x SIZE %d\n",a
       ,chunk,NEXTPIECE(chunk),PREVPIECE(chunk),SIZEPIECE(chunk));
}
*/

void _dccinitmem()
/* Initialize (non-Btree, slow form) */
{ firstchunk=(char *)(((int)sbrk(0)+3)&0xfffffffc);
  if(brk(firstchunk+SBRKSZ))
    LBERROR((stderr,"Can't even get %d memory!\n",SBRKSZ),exit(-1));
  maxposition=sbrk(0);
  SIZEPIECE(firstchunk)=HEADSZ;
  NEXTPIECE(firstchunk)=NULL;
  spacefound=MALLOC_BASIC;
  lastchunk=firstchunk;
}

void free();
char *malloc();

static char *_dccmalloc(sz) int sz;
/* malloc() that doesn't use Btree.  This is included so that
 * stuff initialized by _dccmain() doesn't cause the large (300 line)
 * Btree junk to be linked into the even the smallest programs. */
{ char *chunk,*startptr,*endptr;
  int cont,fullsize;

  if(!spacefound) _dccinitmem();

  sz=((sz+(HEADSZ+3))&0xfffffffc);
  for(chunk=firstchunk,cont=1;cont;chunk=NEXTPIECE(chunk))
  { startptr=chunk+SIZEPIECE(chunk);
    endptr=NEXTPIECE(chunk);
    if(!endptr)
    { cont=0; endptr=maxposition; }
    fullsize=endptr-startptr;
    if(fullsize<0) return NULL;
    if(fullsize>=sz)
    { NEXTPIECE(startptr)=NEXTPIECE(chunk);
      NEXTPIECE(chunk)=startptr;
      if(NEXTPIECE(startptr)!=NULL) PREVPIECE(NEXTPIECE(startptr))=startptr;
      else lastchunk=startptr;
      PREVPIECE(startptr)=chunk;
      SIZEPIECE(startptr)=sz;
     return startptr+HEADSZ;
    }
  }
 return NULL;
}

   /* Btree stuff for malloc() */

#define N        32
#define SIZE     (N*2)
#define MININSSZ HEADSZ

typedef struct memnode NODE;

struct node_ele
{ int space;
  char *pptr;
  NODE *child;
};

struct memnode
{ int sz;
  struct node_ele elem[SIZE+2];
};
/* elem[0].child,elem[0].stf,...,elem[sz-1].stuff,elem[sz].child */

static int memdepth;
static NODE *memtree;

#define MAXDEPTH 7 /* actually, 6 + 1 spare */

static int nodecnt,neednodecnt;
NODE **topnode;
static char newmalloc_disallow;

static NODE *nodealloc()
{ NODE *ptr;
  if(nodecnt)
  { nodecnt--;
    ptr=(NODE *)topnode;
    topnode=(NODE **)*topnode;
   return ptr;
  }
  else
    LBERROR((stderr,"ERR nodealloc() memdepth %d nodecnt %d neednodecnt %d\n"
      ,memdepth,nodecnt,neednodecnt),return NULL);
}

static void nodefree(ptr) NODE *ptr;
{ nodecnt++;
  *(NODE ***)ptr=topnode;
  topnode=(NODE **)ptr;
}

static int alloc_cnt;

#define adjustneednodecnt() (neednodecnt=memdepth*2+MAXDEPTH+(alloc_cnt/64))

static void checknodecnt()
{ static adjustingnode;
  NODE *ptr;

  if(!adjustingnode)
  { adjustingnode=1;
    newmalloc_disallow=0;
    do
    { adjustneednodecnt();
      while(nodecnt<neednodecnt) /*Reset space */
      { ptr=(NODE *)malloc(sizeof *ptr);
        if(!ptr)
        { adjustingnode=0;
          newmalloc_disallow=1;
         return;
        }
        nodefree(ptr); /* Add this ptr to pool. */
      }
      while(nodecnt>neednodecnt+MAXDEPTH) /*Reset space */
      { ptr=nodealloc(); /* retreive ptr from pool. */
        if(!ptr)
        { adjustingnode=0;
          newmalloc_disallow=1;
         return;
        }
        free(ptr);
      }
    } while(nodecnt<neednodecnt); /* no way it could be > */
    adjustingnode=0;
  }
}

static void Bcreate(tree) NODE **tree;
/* Be very carefull that _dccmalloc() and _dccinitmem()
 * don't allocate too many items to put in 1 node of memtree */
{ int i;
  (*tree)=(NODE *)nodealloc();
  if(!(*tree))
  { printf("can't get enough memory\n");
   exit(-1);
  }
  (*tree)->sz=0;
  (*tree)->elem->child=NULL;
}

static int Bbinsearch(elem,sz,space,pptr)
   struct node_ele *elem; int sz,space; char *pptr;
{ int low,high,mid;
  struct node_ele *ptr;

  low=0; high=sz-1;

  while(1)
  { if(low>high) break;
    mid=(low+high)/2;
    ptr=elem+mid;
    if(ptr->space<space) low=mid+1;
    else if(ptr->space>space) high=mid-1;
    else if(ptr->pptr<pptr) low=mid+1;
    else if(ptr->pptr>pptr) high=mid-1;
    else /* they are equal */
      return mid;
  }
 return -1-low;
}

static int Binsert(tree,parent,psplit,space,pptr)
    NODE **tree,*parent; int psplit,space; char *pptr;
{ int split,success,newparent;
  NODE *temp;

  if(!pptr) LBERROR((stderr,"ERR Binsert() NULL pptr\n"),return 0);
     /* NULL pointers mess up deletion routine */

  if(space<MININSSZ) return 1;

  split=Bbinsearch((*tree)->elem,(*tree)->sz,space,pptr);
  if(split>=0)
     LBERROR((stderr,"ERR Binsert() already inserted space %d pptr %d\n"
        ,space,pptr),return 0);
  split= -1-split;

  if(!(*tree)->elem->child)  /* if at bottom of tree insert info directly */
  { success=1;

    memmove((*tree)->elem+split+1,(*tree)->elem+split
       ,((*tree)->sz-split+1)*sizeof(*(*tree)->elem));

    (*tree)->elem[split].space=space;
    (*tree)->elem[split].pptr=pptr;
    (*tree)->sz++;
  }
  else success=Binsert(&(*tree)->elem[split].child,(*tree),split,space,pptr);

  if((*tree)->sz>SIZE)                      /* keep within size bounds */
  { newparent= !parent;             /* grow upwards if necessary */
    if(newparent)
    { Bcreate(&parent);
      memdepth++;
    }

    memmove(parent->elem+psplit+1,parent->elem+psplit /* make room */
       ,(parent->sz-psplit+1)*sizeof(*parent->elem));

    temp=(*tree); /* *tree may be changed by below */
    parent->elem[psplit]=temp->elem[N]; /* insert splitting element */
    parent->elem[psplit].child=temp;
    Bcreate(&parent->elem[psplit+1].child);
    parent->sz++;

    temp=parent->elem[psplit+1].child;           /* do the split */
    memcpy(temp->elem,(*tree)->elem+N+1,(N+1)*sizeof(*temp->elem));

    temp->sz=(*tree)->sz=N;

    if(newparent) (*tree)=parent;     /* acknowledge any upword growth */
  }
 return success;
}

static void Bfixtree(tree,parent,split)
    NODE *tree; NODE **parent; int *split;
/* Only called if tree->sz is smaller then N.  This procedure combines
 * elements of parent and tree's siblings so that every node at tree's depth
 * has at least N elements.  parent and split are adjusted to point to the
 * new location of that element. */
{ NODE *temp;

  if(*split&&(temp=(*parent)->elem[*split-1].child)->sz>N)
     /* if can borrow an item from the left, do a right shift */
  { memmove(tree->elem+1,tree->elem,(tree->sz+1)*sizeof(*temp->elem));
    *tree->elem=(*parent)->elem[*split-1];
    tree->elem->child=temp->elem[temp->sz].child;
    temp->sz--;

    (*parent)->elem[*split-1].space=temp->elem[temp->sz].space;
    (*parent)->elem[*split-1].pptr=temp->elem[temp->sz].pptr;
    tree->sz++;
  }
  else if(*split<(*parent)->sz&&(temp=(*parent)->elem[*split+1].child)->sz>N)
     /* if can borrow an item from the right, do a left shift */
  { tree->elem[tree->sz].space=(*parent)->elem[*split].space;
    tree->elem[tree->sz].pptr=(*parent)->elem[*split].pptr;
    tree->elem[++tree->sz].child=temp->elem->child;

    (*parent)->elem[*split].space=temp->elem->space;
    (*parent)->elem[*split].pptr=temp->elem->pptr;

    memmove(temp->elem,temp->elem+1,temp->sz*sizeof(*temp->elem));
    temp->sz--;

    *parent= tree;
    *split=tree->sz-1;
  }
  else /* can't borrow individual element from either side. */
  { int newsplit;
    if(*split<(*parent)->sz)
       /* if can combine with the tree on the right, do so. */
    { temp=(*parent)->elem[*split+1].child;
      tree->elem[tree->sz].space=(*parent)->elem[*split].space;
      tree->elem[tree->sz].pptr=(*parent)->elem[*split].pptr;
      newsplit=tree->sz;
      tree->sz++;

      memcpy(tree->elem+tree->sz,temp->elem,(temp->sz+1)*sizeof(*temp->elem));
      tree->sz+=temp->sz;
      nodefree(temp);/********  ****** */

      memmove((*parent)->elem+ *split,(*parent)->elem+ *split+1
        ,((*parent)->sz- *split)*sizeof(*temp->elem));
      (*parent)->elem[*split].child= tree;
      (*parent)->sz--;

      *parent= tree;
      *split=newsplit;
    }
    else /* *split==(*parent)->sz
          * logically, it must be possible to combine with left; so do it. */
    { if(!*split)
      { fputs("Internal error in Bfixtree() (for malloc)",stderr);
       exit(-1);
      }
      temp=(*parent)->elem[*split-1].child;
      memmove(tree->elem+(1+temp->sz),tree->elem
        ,(tree->sz+1)*sizeof(*temp->elem));

      memcpy(tree->elem,temp->elem,(temp->sz+1)*sizeof(*temp->elem));

      tree->elem[temp->sz].space=(*parent)->elem[*split-1].space;
      tree->elem[temp->sz].pptr=(*parent)->elem[*split-1].pptr;
      tree->sz+=1+temp->sz;

      nodefree(temp);/********  ****** */

      (*parent)->elem[--(*parent)->sz].child=tree;
    }
  }
}

static int Bdelete_xchg(tree,parent,psplit,ele)
    NODE *tree,**parent; int *psplit; struct node_ele *ele;
/* removes and returns the pointer associated with the largest element in
 * the tree.  Also adjusts each level of the tree with Bfixtree().
 * parant, psplit- start out pointing to split element that pointed to this
 * node.  If fixtree changes things, these are set to point at the new
 * location of that node. */
{ NODE *temp;
  int split;
  int stat;

  if(!tree) return 0;

  temp=tree;
  split=temp->sz;
  stat=Bdelete_xchg(temp->elem[split].child,&temp,&split,ele);
  if(!stat) *ele=tree->elem[--tree->sz];
  if(tree->sz<N&&parent) Bfixtree(tree,parent,psplit);
 return 1;
}

static char *Bdelete(comtree,parent,psplit,space,pptr)
     NODE **comtree,**parent; int psplit,space; char *pptr;
/* Deletes the item with smallest space>space or (if small_space==space)
 *  it deletes the one with the smallest pptr */
{ int split;
  NODE *temp,*tree;
  char *stat;

  if(space<MININSSZ) return pptr;

  tree= *comtree;

  if(!tree) return NULL;
  split=Bbinsearch(tree->elem,tree->sz,space,pptr);

  if(split>=0) /* exact match, can't get any better */
  { stat=NULL;
  }
  else
  { split= -1-split;
    stat=Bdelete(&tree->elem[split].child,comtree,split,space,pptr);
  }

  if(!stat&&split<tree->sz&&tree->elem[split].space>=space)
  { int rplc;
    struct node_ele nd;

    temp=tree;
    rplc=Bdelete_xchg(tree->elem[split].child,&temp,&split,&nd);

    stat=temp->elem[split].pptr;
    if(rplc)
    { temp->elem[split].space=nd.space;
      temp->elem[split].pptr=nd.pptr;
      if(tree->sz==0)
      { nodefree(tree);
        memdepth--;
        (*comtree)=tree=temp;
      }
    }
    else /* we were already at the bottom */
    { memmove(temp->elem+split,temp->elem+split+1
         ,(temp->sz-split+1)*sizeof(*temp->elem));
      temp->sz--;
    }
  }

  if(tree->sz<N&&parent)
  { temp= *parent;
    Bfixtree(tree,&temp,&psplit);
    if((*parent)->sz==0)
    { nodefree(*parent);
      (*parent)=tree;
      memdepth--;
    }
  }
 return stat;
}

#ifdef UNDEF
void _dcc_checktree(tree,depth)
   NODE *tree; int depth;
{ int i;
  static cnt;
  static firstnode,lastspace;
  static char *lastpptr;

  if(!depth) firstnode=1;

  if(tree)
  { for(i=0;i<tree->sz;i++)
    { _dcc_checktree(tree->elem[i].child,depth+1);
      if(!firstnode)
      { if(tree->elem[i].space<lastspace||(tree->elem[i].space==lastspace
          &&tree->elem[i].pptr<=lastpptr)) fprintf(stderr,"INCONSISTENT\n");
      }
      else firstnode=0;
/*
      fprintf(stderr,"%*d %d\n",70-10*depth,tree->elem[i].space
         ,tree->elem[i].pptr);
*/
      lastspace=tree->elem[i].space;
      lastpptr=tree->elem[i].pptr;
    }
    _dcc_checktree(tree->elem[i].child,depth+1);
  }
  else if(depth-1!=memdepth)
    fprintf(stderr,"INCONSISTENT2 %d %d\n",depth,memdepth);
}
#endif /*UNDEF*/

char *malloc(sz) int sz;
{ char *chunk,*startptr,*endptr;
  int cont,fullsize;

  if(newmalloc_disallow) return NULL;

  if(!(spacefound&MALLOC_BTREE))
  { if(!spacefound) _dccinitmem();
    nodefree(_dccmalloc(sizeof(NODE)));
    Bcreate(&memtree);
    for(chunk=firstchunk,cont=1;cont;chunk=NEXTPIECE(chunk))
    { startptr=chunk+SIZEPIECE(chunk);
      endptr=NEXTPIECE(chunk);
      if(!endptr)
      { cont=0; endptr=maxposition; }
      fullsize=endptr-startptr;
      if(fullsize<0) LBERROR((stderr,"ERR malloc() 3\n"),return NULL);
      if(!Binsert(&memtree,NULL,0,fullsize,chunk)) return NULL;
    }
    spacefound=MALLOC_BTREE;
  }
  sz=((sz+(HEADSZ+3))&0xfffffffc);
  chunk=Bdelete(&memtree,NULL,0,sz,NULL);
  if(!chunk)
  { checknodecnt();
    if(NEXTPIECE(lastchunk)!=NULL)
       LBERROR((stderr,"ERR malloc 3.4\n"),return NULL);

      /* extend brk address */
    if(brk(lastchunk+SIZEPIECE(lastchunk)+sz+SBRKSZ))
      return NULL; /* Standard out-of-memory: */
    if(Bdelete(&memtree,NULL,0,maxposition-(lastchunk+SIZEPIECE(lastchunk))
           ,lastchunk)!=lastchunk)
      LBERROR((stderr,"ERR malloc() 3.5\n"),return NULL);
    if(!Binsert(&memtree,NULL,0,sz+SBRKSZ,lastchunk))
      LBERROR((stderr,"ERR malloc() 3.6\n"),return NULL);
    maxposition=sbrk(0);

    chunk=Bdelete(&memtree,NULL,0,sz,NULL);  /* now try again */
    if(!chunk)
    { checknodecnt();
      LBERROR((stderr,"ERR malloc() 3.7\n"),return NULL);
    }
  }
  
  startptr=chunk+SIZEPIECE(chunk);
  endptr=NEXTPIECE(chunk);
  if(!endptr) endptr=maxposition;
  fullsize=endptr-startptr;

  if(fullsize<sz) LBERROR((stderr,"ERR malloc() 2\n"),return NULL);
  if(!Binsert(&memtree,NULL,0,fullsize-sz,startptr))
    LBERROR((stderr,"ERR malloc() 1\n"),return NULL);

  NEXTPIECE(startptr)=NEXTPIECE(chunk);
  NEXTPIECE(chunk)=startptr;
  if(NEXTPIECE(startptr)!=NULL) PREVPIECE(NEXTPIECE(startptr))=startptr;
  else lastchunk=startptr;
  PREVPIECE(startptr)=chunk;
  SIZEPIECE(startptr)=sz;

  checknodecnt();
  alloc_cnt++;
 return startptr+HEADSZ;
}

char *realloc(ptr,sz) char *ptr;
{ char *chunk,*startptr,*endptr,*prevchunk;
  int fullsize;
  int oldsz;

  if(!ptr) return malloc(sz);

  if(!spacefound) LBERROR((stderr,"ERR spacefound realloc()\n"),return NULL);
  sz=((sz+(HEADSZ+3))&0xfffffffc);
  chunk=ptr-HEADSZ;
  prevchunk=PREVPIECE(chunk);
  startptr=prevchunk+SIZEPIECE(prevchunk);

  endptr=NEXTPIECE(chunk);
  if(endptr==NULL) endptr=maxposition;

  oldsz=SIZEPIECE(chunk);
  fullsize=endptr-startptr;
  if(fullsize<0)
    LBERROR((stderr
      ,"ERR realloc() 3 fullsize %d endptr %#x startptr %#x\n",fullsize,endptr
      ,startptr),return NULL);
  if(fullsize>=sz)
  { if(Bdelete(&memtree,NULL,0,endptr-(chunk+SIZEPIECE(chunk)),chunk)!=chunk)
      LBERROR((stderr,"ERR realloc() 1\n"),return NULL);
    if(Bdelete(&memtree,NULL,0,chunk-startptr,prevchunk)!=prevchunk)
      LBERROR((stderr,"ERR realloc() 2\n"),return NULL);

    if(!Binsert(&memtree,NULL,0,fullsize-sz,startptr))
      LBERROR((stderr,"ERR realloc() 4\n"),return NULL);
    if(startptr!=chunk)
    { if(oldsz<sz)
        memmove(startptr,chunk,oldsz);
      else
        memmove(startptr,chunk,sz);
      NEXTPIECE(prevchunk)=startptr;
      if(NEXTPIECE(startptr)!=NULL) PREVPIECE(NEXTPIECE(startptr))=startptr;
      else lastchunk=startptr;
      chunk=startptr;
    }
    SIZEPIECE(chunk)=sz;

    checknodecnt();
   return chunk+HEADSZ;
  }
  else
  { sz-=HEADSZ; chunk+=HEADSZ;
    startptr=malloc(sz);
    if(startptr!=NULL)
    { if(oldsz<sz)
        memcpy(startptr,chunk,oldsz);
      else
        memcpy(startptr,chunk,sz);
      free(chunk);
    }
   return startptr;
  }
}

void free(ptr) char *ptr;
{ char *chunk,*startptr,*endptr,*prevchunk,*ret;
  int fullsize;

  if(!ptr) return;

  if(!spacefound) LBERROR((stderr,"ERR spacefound in free()\n"),return);
  chunk=ptr-HEADSZ;
  prevchunk=PREVPIECE(chunk);
  startptr=prevchunk+SIZEPIECE(prevchunk);

  endptr=NEXTPIECE(chunk);
  if(endptr==NULL)
  { endptr=maxposition;
    lastchunk=prevchunk;
  }

  if((ret=Bdelete(&memtree,NULL,0,endptr-(chunk+SIZEPIECE(chunk)),chunk))
      !=chunk)
    LBERROR((stderr,"ERR free() 1 sz %d chunk %d ret %d\n"
      ,endptr-(chunk+SIZEPIECE(chunk)),chunk,ret),return);
  if((ret=Bdelete(&memtree,NULL,0,chunk-startptr,prevchunk))!=prevchunk)
    LBERROR((stderr,"ERR free() 2 sz %d prevchunk %d ret %d\n"
      ,chunk-startptr,prevchunk,ret),return);

  fullsize=endptr-startptr;
  if(fullsize<0) LBERROR((stderr,"ERR free() 2.5\n"),return);
  if(!Binsert(&memtree,NULL,0,fullsize,prevchunk))
    LBERROR((stderr,"ERR free() 3\n"),return);

  NEXTPIECE(prevchunk)=NEXTPIECE(chunk);
  if(NEXTPIECE(prevchunk)!=NULL) PREVPIECE(endptr)=prevchunk;

  alloc_cnt--;
  checknodecnt();
}
