/* mtst2.c - memory allocation tester type 2
 * (do more then just boundary check) */

#include <stdio.h>

#define HEADSZ    32      /* space between allocs for overflow detect */
#define AREASZ    1000000 /* total heap size */
#define MAX_COUNT 10000   /* maximum # of allocated objects */

#define CONSIST_COUNT 10000
#define CHECKCHAR     '~'

#ifndef NULL
# define NULL 0
#endif

#define MSG(x) do                                \
    { fprintf x;                                 \
      fprintf(stderr,":\n\7 (count %d callno %d)\7\7\7\n",count,callno); \
      sleep(5);                                  \
    } while(0)

#define FMSG(x) do                               \
    { fprintf x;                                 \
      fprintf(stderr,":\n\7 (count %d callno %d)\7\7\7\n",count,callno); \
     exit(-1);                                   \
    } while(0)

static char *area;
static int count;
static struct places
  { int off;
    int sz;
  } *place;
static int inited;

char *malloc();
static void checkspecific();

static int callno;
static int con_entry,con_off;

static int binsearch(off) int off;
{ int mid,low,high,thisoff;
  low=0;
  high=count-1;
  while(1)
  { if(low>high)
    { fprintf(stderr,"thisoff %d mid %d\n",thisoff,mid);
      for(mid=0;mid<count;mid++)
        fprintf(stderr,"[%3d,%6d]        ",mid,place[mid].off);
      FMSG((stderr
       ,"internal inconsistency in mtst2, binsearch off %d low %d high %d"
       ,off,low,high));
    }
    thisoff=place[mid=(low+high)/2].off;
    if(thisoff<off) low=mid+1;
    else if(thisoff>off) high=mid-1;
    else return mid;
  }
}

static void checkconsistency()
{ int to_do;

  { static lastline;
    extern yylineno;

    if(yylineno!=lastline)
    { fprintf(stderr,"yylineno %d count %d\n",yylineno,count);
      lastline=yylineno;
    }
  }
/* */

  callno++;

  if(con_off<place[con_entry-1].off+place[con_entry-1].sz)
     con_off=place[con_entry-1].off+place[con_entry-1].sz;
  to_do=CONSIST_COUNT;
  while(1)
  { if(con_entry>=count) con_entry=1,con_off=0;
    while(con_off>=place[con_entry].off)
    { con_off=place[con_entry].off+place[con_entry].sz;
      con_entry++;
      if(con_entry>=count) con_entry=1,con_off=0;
    }
    if(con_off+to_do<=place[con_entry].off)
    { checkspecific(con_off,to_do,"checkconsistency");
     return;
    }
    else
    { int size;

      size=place[con_entry].off-con_off;
      to_do-=size;
      checkspecific(con_off,to_do,"checkconsistency2");
    }
  }
}

static void checkbound(entry) int entry;
{ checkspecific(place[entry].off-HEADSZ,HEADSZ,"checkbound");
  checkspecific(place[entry].off+place[entry].sz,HEADSZ,"checkbound2");
}

typedef int bulk_CHECK_TP;
static bulk_CHECK_TP bulk_CHECKCHAR;

static void checkspecific(off,sz,str) int off,sz; char *str;
{ int a;
  char *ptr,*stopptr;

  a=sz;
  ptr=area+off;

    /* PORT: from here blank line */
  for(;a>0&&(((int)ptr)&0x3);a--,ptr++) /* align on dword boundary */
  { if(*ptr!=CHECKCHAR)
       MSG((stderr,"%s illegally written memory off %d (alignment)",str
          ,ptr-area));
  }
  stopptr=ptr+(a-4);
  for(;ptr<=stopptr;ptr+=sizeof(bulk_CHECK_TP))
  { if(*(bulk_CHECK_TP *)ptr!=bulk_CHECKCHAR)
     MSG((stderr
     ,"%s illegally written memory off %d value %#x b_off %d sz %d(bulk check)"
     ,str,ptr-area,*(bulk_CHECK_TP *)ptr,off,sz));
  }
  a&=0x3;
    /* PORT: end nonportable area */

  for(;a>0;a--,ptr++) /* get any left over */
  { if(*ptr!=CHECKCHAR)
       MSG((stderr,"%s illegally written memory off %d b_off %d sz %d"
        ,str,ptr-area,off,sz));
  }
}

static void setspecific(off,sz) int off,sz;
{ int a,stop;

  for(a=off,stop=off+sz;a<stop;a++) area[a]=CHECKCHAR;
}

static void removeplace(entry) int entry;
{ if(con_entry>entry) con_entry--;
    /* if == it will just go to the next one rather then removed */

  count--;
  if(count<2)
     FMSG((stderr,"internal inconsistency; count=%d too small",count));
  memmove(place+entry,place+entry+1,(count-entry)*sizeof(*place));
}

static void insertplace(entry,off,sz) int entry,off,sz;
{ if(con_entry>entry) con_entry++;
    /* if == it will either go to the insertion, or it will be forwarded
     *   to go to current con_entry */

  memmove(place+entry+1,place+entry,(count-entry)*sizeof(*place));
  place[entry].off=off; place[entry].sz=sz;
  if(count>=MAX_COUNT)
     FMSG((stderr,"count exceeds limit of %d",MAX_COUNT));
  count++;
}

char *testmalloc(size) int size;
{ int usesize;
  int a;
  int chunkoff;

  if(!inited)
  { area=malloc(AREASZ);
    place=(struct places *)malloc(MAX_COUNT*sizeof(*place));
    if(!area||!place) FMSG((stderr,"Can't init mtst2"));

    setspecific(0,AREASZ);

    place[0].off=0; place[0].sz=0;
    place[1].off=AREASZ-HEADSZ; place[1].sz=0;
    count=2;

    for(a=0;a<sizeof(bulk_CHECKCHAR);a++)  /* PORT: somewhat non-portable */
      ((char *)&bulk_CHECKCHAR)[a]=CHECKCHAR;

    inited=1;
  }

  checkconsistency();

  usesize=(size+(2*HEADSZ+3))&0xfffffffc;
  for(a=count-2;a>=0;a--)
  { int chunksize;

    chunksize=place[a+1].off-place[a].off-place[a].sz;
    if(usesize<=chunksize)
    { a++;
      chunkoff=place[a].off-(usesize-HEADSZ);
      insertplace(a,chunkoff,size);
      checkspecific(chunkoff,size,"testmalloc");
      checkbound(a);
     return area+chunkoff;
    }
  }
 return NULL;
}

void testfree(ptr) char *ptr;
{ int entry;

  checkconsistency();
  entry=binsearch(ptr-area);
  checkbound(entry);
  setspecific(place[entry].off,place[entry].sz);
  removeplace(entry);
}

char *testrealloc(ptr,size) char *ptr; int size;
{ char *outptr;
  int entry;
  int chunksize;
  int usesize;
  int chunkoff;

  usesize=(size+(2*HEADSZ+3))&0xfffffffc;
  entry=binsearch(ptr-area);
  chunksize=place[entry+1].off-(place[entry-1].off+place[entry-1].sz);
  if(usesize<=chunksize)
  { int placeend;

    checkconsistency();

    chunkoff=place[entry+1].off-(usesize-HEADSZ);

    checkbound(entry);
    placeend=place[entry].off+place[entry].sz;
    if(place[entry].off>chunkoff)
      checkspecific(chunkoff,place[entry].off-chunkoff,"realloc");
    if(chunkoff+size>placeend)
      checkspecific(placeend,(chunkoff+size)-placeend,"realloc2");
    memmove(area+chunkoff,area+place[entry].off
         ,(size<place[entry].sz?size:place[entry].sz));

    if(place[entry].off<chunkoff)  /* set low chunk */
    { if(placeend>chunkoff)
         setspecific(place[entry].off,chunkoff-place[entry].off);
      else
         setspecific(place[entry].off,place[entry].sz);
               /*placeend-place[entry].off*/
    }
    if(chunkoff+size<placeend)    /* set high chunk */
    { if(chunkoff+size>place[entry].off)
        setspecific(chunkoff+size,placeend-(chunkoff+size));
      else
        setspecific(place[entry].off,place[entry].sz);
    }

    place[entry].off=chunkoff;
    place[entry].sz=size;
   return area+chunkoff;
  }
  else
  { int copysize;

    outptr=testmalloc(size);
    if(!outptr) return NULL;

      /* compensate for malloc moving entries */
    if(place[entry].off<ptr-area) entry++;

    copysize=size<=place[entry].sz?size:place[entry].sz;
    memcpy(outptr,ptr,copysize);
    testfree(ptr);
   return outptr;
  }
}
