/*maltest.c - random malloc tester */
#include <stdio.h>
#include <alloc.h>

#define KEYITR 56780

/*#define ITRTEST
*/

#ifdef ITRTEST
# define ITRPRINT(args) if(i>KEYITR) fprintf args
#else
# define ITRPRINT(args)
#endif

#define STUFFSZ 10000 /* max # of items malloced at one time in test */

struct
 { char *ptr;
   char sz,fill;
 } stuff[STUFFSZ];
int i;       /* iterations through main loop */
int obj_c;   /* # of currently malloc'ed items (by manual count) */
int ml_c;    /* # of calls to malloc() */
int rl_c;    /* # of calls to realloc() */
int f_c;     /* # of calls to free() */
int afill_c; /* # of times everything has been malloc'ed */
int aclr_c;  /* # of times everything has been free'd */
int mclr_c;  /* # of times many have been cleared*/
int start_time;
char *keyptr;

void printstat()
{ fprintf(stderr,"On iteration %d: %d objects, %d malloc calls,\n",i,obj_c
    ,ml_c);
  fprintf(stderr,"  %d realloc calls, %d free calls,\n",rl_c,f_c);
  fprintf(stderr,"  all filled %d, all cleared %d\n",afill_c,aclr_c);
  fprintf(stderr,"  many cleared %d, Time %.2f\n",mclr_c
    ,(timer()-start_time)/18.2);
}

void error(char *str)
{ fprintf(stderr,"Fatal error in %s\n",str);
  printstat();
 exit(-1);
}

void check_fill(int itm,char *errmsg)
{ int a;
  char *ptr;
  char fill;

  fill=stuff[itm].fill;
  if(!stuff[itm].ptr) error(errmsg);
  for(a=stuff[itm].sz-1,ptr=stuff[itm].ptr;a>=0;a--,ptr++)
  { if(*ptr!=fill)
    { printf("Error: fill %d *ptr %d itm %d a %d sz %d\n",fill,*ptr,itm,a
        ,stuff[itm].sz);
      error("fill stuff");
    }
  }
}

void do_free(int itm)
{ if(!stuff[itm].ptr) error("do_free()");
  check_fill(itm,"do_free() on check_fill() problem");
  free(stuff[itm].ptr);
  stuff[itm].ptr=0;
  f_c++; obj_c--;

ITRPRINT((stderr," DO_FREE "));
}

void do_malloc(int itm)
{ int a;
  char fill,*ptr;
  if(stuff[itm].ptr) error("do_malloc()");

  stuff[itm].sz=rand()%100+1;
  stuff[itm].ptr=malloc(stuff[itm].sz);
  if(!stuff[itm].ptr) error("do_malloc() 2");

  stuff[itm].fill=fill=rand(); /* auto-truncate to 8 bits (or whatever) */
  for(a=stuff[itm].sz-1,ptr=stuff[itm].ptr;a>=0;a--,ptr++) *ptr=fill;
  ml_c++; obj_c++;
ITRPRINT((stderr," DO_MALLOC "));
}

void do_realloc(int itm)
{ int sz,a;
  char *ptr,fill;

  if(!stuff[itm].ptr) error("do_realloc()");
  check_fill(itm,"do_realloc() on check_fill() problem");

  sz=rand()%100+1;
  ptr=realloc(stuff[itm].ptr,sz);
  if(!ptr) error("do_realloc() 2");
  stuff[itm].ptr=ptr;
  if(sz<stuff[itm].sz) stuff[itm].sz=sz;

  check_fill(itm,"do_realloc() on check_fill() problem 2 ");

  stuff[itm].sz=sz;

  stuff[itm].fill=fill=rand(); /* auto-truncate to 8 bits (or whatever) */
  for(a=stuff[itm].sz-1,ptr=stuff[itm].ptr;a>=0;a--,ptr++) *ptr=fill;
  rl_c++;

ITRPRINT((stderr," DO_REALLOC "));
}

main(int argc,char *argv[])
{ int a,b;

  if(argc<2||!(a=atoi(argv[1])))
  { error("\nUsage: programname non-0-randomseed");
  }

  if(argc>=3) keyptr=atoi(argv[2]);

  srand(a);
  start_time=timer();

  while(!iskey()||getch()!='q')
  { if(!(i%1000)) printstat();
    i++;
    a=rand();

ITRPRINT((stderr,"i %d a %d a%%40000 %d\n",i,a,a%40000));
    switch(a%40000)
    { case 0: /* free() everything */
       for(a=0;a<STUFFSZ;a++)
       { if(stuff[a].ptr) do_free(a);
       }
       aclr_c++;
      break;
      case 1: /* malloc() everything */
       for(a=0;a<STUFFSZ;a++)
       { if(!stuff[a].ptr) do_malloc(a);
       }
       afill_c++;
      break;
      case 2: /* conditionally free everything */
       for(a=0;a<STUFFSZ;a++)
       { if(stuff[a].ptr&&rand()%2) do_free(a);
       }
       mclr_c++;
      break;
      default:
       a%=STUFFSZ;
       if(stuff[a].ptr) /* allocated? */
       { if(rand()%2) do_realloc(a);
         else do_free(a);
       }
       else /* not allocated */
         if(rand()%2) do_malloc(a);
    }
  }
  printstat();
}
